/*========================================================*/
/*                                                        */
/*  fit2rotgauss.c      version 0.8.1     2010.01.15      */
/*                                                        */
/*  Copyright (C) 2010 by Wojtek Pych, CAMK PAN           */
/*  pych@camk.edu.pl                                      */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*                                                        */
/*  Make a nonlinear least squares fit of                 */
/*  2 rotational profiles convolved with a gaussian.      */
/*                                                        */
/*  The rotational profile defined as:                    */
/*                                                        */
/*  y = A [ (1 - B) sqrt(z) + Pi/4 B z ] + E              */
/*  where z = (1 - a*a); a = (v - C)/D                    */
/*  A: amplitude                                          */
/*  B: linear limb darkening coefficient                  */
/*  C: radial velocity of the center of mass              */
/*  D: Vsini (rotation speed)                             */
/*                                                        */
/*  Gaussian defined as:                                  */
/*                                                        */
/*  y = 1/sqrt(Pi)/S * exp[ -(i-n/2)*(i-n/2)/S/S ]        */
/*                                                        */
/*========================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>

#include "pfitsio.h"

struct data
{
  int     n;
  double  *t,
          *y,
          *sigma;
};

typedef struct
{
  short   va1, vb1, vc1, vd1,
          va2, vb2, vc2, vd2,
          ve, vs,
          verbose;
  int     max_iter;
  double  a1, b1, c1, d1, a2, b2, c2, d2, e, s,
          epsabs,
          epsrel,
          lim;
} PARAMS;

PARAMS  par;

/* Control values */
  const size_t                      p = 10;   // number of fit parameters
  const gsl_multifit_fdfsolver_type *T;
  const gsl_rng_type                *type;

#define STRLEN 200

void    usage(void);
void    read_par(char *);
double  *rotprof(int, double *, double, double, double, double);
double  *gaussprof(int, double);
double  *convolve(int, double *, int, double *);
int     expb_f(const gsl_vector *, void *, gsl_vector *);
int     expb_df(const gsl_vector *, void *, gsl_matrix *);
int     expb_fdf(const gsl_vector *, void *, gsl_vector *, gsl_matrix *);
void    print_state(int, gsl_multifit_fdfsolver *, double);

/*--------------------------------------------------------*/
void usage()
{
  printf("\n\t USAGE:  fit2rotgauss  input_file  parameters_file\n\n");
  exit(1);
}
/*--------------------------------------------------------*/
void read_par(char *parfname)
{
          char  line[STRLEN],
                key[STRLEN],
                val[STRLEN];
          FILE  *inpf;

  if ((inpf=fopen(parfname, "r")) == NULL)
  { perror(parfname); exit(EXIT_FAILURE); }

  while (feof(inpf) == 0)
  {
    if (fgets(line, STRLEN, inpf) == NULL)
    { perror("\n\tERROR! read_par()\n"); exit(EXIT_FAILURE); }
    if (feof(inpf) != 0) break;

    sscanf(line, "%s = %s", key, val);

    if (strcasecmp(key, "END") == 0) break;
    else if (strcasecmp(key, "A1") == 0)      sscanf(val, "%lf", &par.a1);
    else if (strcasecmp(key, "B1") == 0)      sscanf(val, "%lf", &par.b1);
    else if (strcasecmp(key, "C1") == 0)      sscanf(val, "%lf", &par.c1);
    else if (strcasecmp(key, "D1") == 0)      sscanf(val, "%lf", &par.d1);
    else if (strcasecmp(key, "A2") == 0)      sscanf(val, "%lf", &par.a2);
    else if (strcasecmp(key, "B2") == 0)      sscanf(val, "%lf", &par.b2);
    else if (strcasecmp(key, "C2") == 0)      sscanf(val, "%lf", &par.c2);
    else if (strcasecmp(key, "D2") == 0)      sscanf(val, "%lf", &par.d2);
    else if (strcasecmp(key, "E") == 0)       sscanf(val, "%lf", &par.e);
    else if (strcasecmp(key, "S") == 0)       sscanf(val, "%lf", &par.s);
    else if (strcasecmp(key, "VAR_A1") == 0)  sscanf(val, "%hd", &par.va1);
    else if (strcasecmp(key, "VAR_B1") == 0)  sscanf(val, "%hd", &par.vb1);
    else if (strcasecmp(key, "VAR_C1") == 0)  sscanf(val, "%hd", &par.vc1);
    else if (strcasecmp(key, "VAR_D1") == 0)  sscanf(val, "%hd", &par.vd1);
    else if (strcasecmp(key, "VAR_A2") == 0)  sscanf(val, "%hd", &par.va2);
    else if (strcasecmp(key, "VAR_B2") == 0)  sscanf(val, "%hd", &par.vb2);
    else if (strcasecmp(key, "VAR_C2") == 0)  sscanf(val, "%hd", &par.vc2);
    else if (strcasecmp(key, "VAR_D2") == 0)  sscanf(val, "%hd", &par.vd2);
    else if (strcasecmp(key, "VAR_E") == 0)   sscanf(val, "%hd", &par.ve);
    else if (strcasecmp(key, "VAR_S") == 0)   sscanf(val, "%hd", &par.vs);
    else if (strcasecmp(key, "EPSABS") == 0)  sscanf(val, "%lf", &par.epsabs);
    else if (strcasecmp(key, "EPSREL") == 0)  sscanf(val, "%lf", &par.epsrel);
    else if (strcasecmp(key, "MAXITER") == 0) sscanf(val, "%d",  &par.max_iter);
    else if (strcasecmp(key, "LIM") == 0)     sscanf(val, "%lf", &par.lim);
    else if (strcasecmp(key, "VERBOSE") == 0) sscanf(val, "%hd", &par.verbose);
    else printf("\tWARNING: Unknown parameter %s\n", key);
  }

  fclose(inpf);

  if (par.verbose > 0)
  {
    printf("Reading: %s\n", parfname);
    printf("-----------\n");
    printf("A1      = %g\n", par.a1);
    printf("B1      = %g\n", par.b1);
    printf("C1      = %g\n", par.c1);
    printf("D1      = %g\n", par.d1);
    printf("A2      = %g\n", par.a2);
    printf("B2      = %g\n", par.b2);
    printf("C2      = %g\n", par.c2);
    printf("D2      = %g\n", par.d2);
    printf("E       = %g\n", par.e);
    printf("S       = %g\n", par.s);
    printf("VAR_A1  = %d\n", par.va1);
    printf("VAR_B1  = %d\n", par.vb1);
    printf("VAR_C1  = %d\n", par.vc1);
    printf("VAR_D1  = %d\n", par.vd1);
    printf("VAR_A2  = %d\n", par.va2);
    printf("VAR_B2  = %d\n", par.vb2);
    printf("VAR_C2  = %d\n", par.vc2);
    printf("VAR_D2  = %d\n", par.vd2);
    printf("VAR_E   = %d\n", par.ve);
    printf("VAR_S   = %d\n", par.vs);
    printf("LIM     = %g\n", par.lim);
    printf("VERBOSE = %d\n", par.verbose);
    printf("-----------\n");
  }
  
  return;
}
/*--------------------------------------------------------*/
double *rotprof(int n, double *t, double a, double b, double c, double d)
{
        int     i;
        double  *y,
                z;

  if ((y=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! rotprof(): calloc(y)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++)
  {
    z=(t[i]-c)/d;
    z=1.0-z*z;
    if (z > 0.0)  y[i]=a*((1.0-b)*sqrt(z)+b*M_PI/4.0*z);
    else          y[i]=0.0;
  }

  return(y);
}
/*--------------------------------------------------------*/
double *gaussprof(int n, double s)
{
        int     i,
                k;
        double  *y,
                *g,
                a,    // normalization
                b;

  if ((y=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! gaussprof(): calloc(y)"); exit(EXIT_FAILURE); }

  k=n/2;
  g=y+k;

  a=sqrt(M_PI)*s;
  for (i=-k; i<=k; i++)
  {
    b=i/s;
    g[i]=exp(-b*b)/a;
  }

  return(y);
}
/*--------------------------------------------------------*/
double *convolve(int nr, double *r, int ng, double *g)
{
        int     i, j,
                k;
        double  *cg,
                *lconv;

  k=ng/2;
  if ((lconv=(double *)calloc(nr, sizeof(double))) == NULL)
  { perror("\n\tERROR! convolve(): calloc(lconv)"); exit(EXIT_FAILURE); }

  cg=g+k;

  for (i=0; i<nr; i++)
  {
    lconv[i]=0.0;
    for (j=-k; j<=k; j++)
    {
      if ((i+j >= 0) && (i+j < nr))  lconv[i]+=r[i+j]*cg[j];
    }
  }

  return(lconv);
}
/*--------------------------------------------------------*/
int expb_f(const gsl_vector *x, void *params, gsl_vector *f)
{
        int     i,
                n,
                ng;
        double  A1, B1, C1, D1,
                A2, B2, C2, D2,
                E, S,
                *t,
                *y,
                *sigma,
                *r, *r1, *r2,
                *g,
                *conv;

  n=((struct data *)params)->n;
  t=((struct data *)params)->t;
  y=((struct data *)params)->y;
  sigma=((struct data *)params)->sigma;

  A1=gsl_vector_get(x, 0);
  B1=gsl_vector_get(x, 1);
  C1=gsl_vector_get(x, 2);
  D1=gsl_vector_get(x, 3);
  A2=gsl_vector_get(x, 4);
  B2=gsl_vector_get(x, 5);
  C2=gsl_vector_get(x, 6);
  D2=gsl_vector_get(x, 7);
  E=gsl_vector_get(x, 8);
  S=fabs(gsl_vector_get(x, 9));

  r1=rotprof(n, t, A1, B1, C1, D1);
  r2=rotprof(n, t, A2, B2, C2, D2);

  if ((r=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_f(): calloc(r)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++) r[i]=r1[i]+r2[i];
  free(r1);
  free(r2);

  ng=2*(int)(S*sqrt(-log(1.0e-8)))+1;
  if (par.verbose > 1) printf("before gaussprof(): S= %f   ng= %d\n", S, ng);
  g=gaussprof(ng, S);

  conv=convolve(n, r, ng, g);

  free(r);
  free(g);

  for (i=0; i<n; i++)
    gsl_vector_set(f, i, (conv[i] + E - y[i])/sigma[i]);

  free(conv);

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
/* Jacobian matrix J(i,j) = dfi / dxj,                              */
/* where  fi = (Yi - yi)/sigma[i],                                  */
/*        z=1-(t-C)*(t-C)/D/D                                       */
/*        Yi = A * [(1-B)*sqrt(z)+B*M_PI/4.0*z] + E                 */
/* and the xj are the parameters (A, B, C, D, E)                    */
/*                                                                  */
/* 0:   df/dA (amplitude)           */
/* 1:   df/dB (limb darkening)      */
/* 2:   df/dC (radial velocity)     */
/* 3:   df/dD (Vsini)               */
/* 4:   df/dE (zero level)          */
/* 5:   df/dS (Gaussian sigma)      */

int expb_df(const gsl_vector *x, void *params, gsl_matrix *J)
{
        int     i,
                n,
                ng;
        double  v, u, z,
                *t,
                *sigma,
                A1, B1, C1, D1,
                A2, B2, C2, D2,
                E, S,
                *g,
                *r, *r1, *r2,
                *conv,
                *dfa1, *dfb1, *dfc1, *dfd1,
                *dfa2, *dfb2, *dfc2, *dfd2,
                *dfe, *dfs, *ldfs;

  n=((struct data *)params)->n;
  t=((struct data *) params)->t;
  sigma=((struct data *) params)->sigma;

  A1=gsl_vector_get(x, 0);
  B1=gsl_vector_get(x, 1);
  C1=gsl_vector_get(x, 2);
  D1=gsl_vector_get(x, 3);
  A2=gsl_vector_get(x, 4);
  B2=gsl_vector_get(x, 5);
  C2=gsl_vector_get(x, 6);
  D2=gsl_vector_get(x, 7);
  E=gsl_vector_get(x, 8);
  S=fabs(gsl_vector_get(x, 9));

  if ((dfa1=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfa1)"); exit(EXIT_FAILURE); }

  if ((dfb1=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfb1)"); exit(EXIT_FAILURE); }

  if ((dfc1=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfc1)"); exit(EXIT_FAILURE); }

  if ((dfd1=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfd1)"); exit(EXIT_FAILURE); }

  if ((dfa2=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfa2)"); exit(EXIT_FAILURE); }

  if ((dfb2=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfb2)"); exit(EXIT_FAILURE); }

  if ((dfc2=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfc2)"); exit(EXIT_FAILURE); }

  if ((dfd2=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfd2)"); exit(EXIT_FAILURE); }

  if ((dfe=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfe)"); exit(EXIT_FAILURE); }

  if ((dfs=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df(): calloc(dfs)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++)
  {
    v=(t[i]-C1)/D1;
    z=1.0-v*v;

    if (z > 0.0)
    {
      if (par.va1)  dfa1[i]=((1.0-B1)*sqrt(z) + B1*M_PI/4.0*z)/sigma[i];
      else          dfa1[i]=0.0;

      if (par.vb1)  dfb1[i]=A1*(M_PI/4.0*z - sqrt(z))/sigma[i];
      else          dfb1[i]=0.0;

      u=((1.0-B1)/sqrt(z) + M_PI/2.0*B1)/sigma[i];

      if (par.vc1) dfc1[i]=A1*v/D1*u;
      else         dfc1[i]=0.0;

      if (par.vd1) dfd1[i]=A1*v*v/D1*u;
      else         dfd1[i]=0.0;
    }
    else
    {
      dfa1[i]=0.0;
      dfb1[i]=0.0;
      dfc1[i]=0.0;
      dfd1[i]=0.0;
    }

    v=(t[i]-C2)/D2;
    z=1.0-v*v;

    if (z > 0.0)
    {
      if (par.va2)  dfa2[i]=((1.0-B2)*sqrt(z) + B2*M_PI/4.0*z)/sigma[i];
      else          dfa2[i]=0.0;

      if (par.vb2)  dfb2[i]=A2*(M_PI/4.0*z - sqrt(z))/sigma[i];
      else          dfb2[i]=0.0;

      u=((1.0-B2)/sqrt(z) + M_PI/2.0*B2)/sigma[i];

      if (par.vc2)  dfc2[i]=A2*v/D2*u;
      else          dfc2[i]=0.0;

      if (par.vd2)  dfd2[i]=A2*v*v/D2*u;
      else          dfd2[i]=0.0;
    }
    else
    {
      dfa2[i]=0.0;
      dfb2[i]=0.0;
      dfc2[i]=0.0;
      dfd2[i]=0.0;
    }

    if (par.ve) dfe[i]=1.0;
    else        dfe[i]=0.0;
  }

  ng=2*(int)(S*sqrt(-log(1.0e-8)))+1;
  if (par.verbose > 2) printf("expb_df(): S= %G  ng= %d\n", S, ng);
  g=gaussprof(ng, S);

  if (par.va1)
  {
    conv=convolve(n, dfa1, ng, g);
    for (i=0; i<n; i++) dfa1[i]=conv[i];
    free(conv);
  }

  if (par.vb1)
  {
    conv=convolve(n, dfb1, ng, g);
    for (i=0; i<n; i++) dfb1[i]=conv[i];
    free(conv);
  }

  if (par.vc1)
  {
    conv=convolve(n, dfc1, ng, g);
    for (i=0; i<n; i++) dfc1[i]=conv[i];
    free(conv);
  }

  if (par.vd1)
  {
    conv=convolve(n, dfd1, ng, g);
    for (i=0; i<n; i++) dfd1[i]=conv[i];
    free(conv);
  }

  if (par.va2)
  {
    conv=convolve(n, dfa2, ng, g);
    for (i=0; i<n; i++) dfa2[i]=conv[i];
    free(conv);
  }

  if (par.vb2)
  {
    conv=convolve(n, dfb2, ng, g);
    for (i=0; i<n; i++) dfb2[i]=conv[i];
    free(conv);
  }

  if (par.vc2)
  {
    conv=convolve(n, dfc2, ng, g);
    for (i=0; i<n; i++) dfc2[i]=conv[i];
    free(conv);
  }

  if (par.vd2)
  {
    conv=convolve(n, dfd2, ng, g);
    for (i=0; i<n; i++) dfd2[i]=conv[i];
    free(conv);
  }

  free(g);

  if (par.vs)
  {
    r1=rotprof(n, t, A1, B1, C1, D1);
    r2=rotprof(n, t, A2, B2, C2, D2);
    if ((r=(double *)calloc(n, sizeof(double))) == NULL)
    { perror("\n\tERROR! expb_df(): calloc(r)"); exit(EXIT_FAILURE); }

    for (i=0; i<n; i++) r[i]=r1[i]+r2[i];
    free(r1);
    free(r2);

    if ((ldfs=(double *)calloc(ng, sizeof(double))) == NULL)
    { perror("\n\tERROR! expb_df(): calloc(ldfs)"); exit(EXIT_FAILURE); }

    for (i=0; i<ng; i++)
    {
      z=(i-ng/2)/S;
      ldfs[i]=2.0*z*z-1.0;
      ldfs[i]*=exp(-z*z)/sqrt(M_PI)/S/S;
    }

    conv=convolve(n, r, ng, ldfs);

    for (i=0; i<n; i++) dfs[i]=conv[i];

    free(ldfs);
    free(conv);
    free(r);
  }
  else
    for (i=0; i<n; i++) dfs[i]=0.0;

  for (i=0; i<n; i++)
  {
    gsl_matrix_set(J, i, 0, dfa1[i]);
    gsl_matrix_set(J, i, 1, dfb1[i]);
    gsl_matrix_set(J, i, 2, dfc1[i]);
    gsl_matrix_set(J, i, 3, dfd1[i]);
    gsl_matrix_set(J, i, 4, dfa2[i]);
    gsl_matrix_set(J, i, 5, dfb2[i]);
    gsl_matrix_set(J, i, 6, dfc2[i]);
    gsl_matrix_set(J, i, 7, dfd2[i]);
    gsl_matrix_set(J, i, 8, dfe[i]);
    gsl_matrix_set(J, i, 9, dfs[i]);
  }

  free(dfa1);
  free(dfb1);
  free(dfc1);
  free(dfd1);
  free(dfa2);
  free(dfb2);
  free(dfc2);
  free(dfd2);
  free(dfe);
  free(dfs);

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
int expb_fdf(const gsl_vector *x, void *params, gsl_vector *f, gsl_matrix *J)
{
  expb_f(x, params, f);
  expb_df(x, params, J);

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
void print_state(int iter, gsl_multifit_fdfsolver *s, double dt)
{
  printf("iter: %3u    |f(x)|= %g\n", iter, gsl_blas_dnrm2(s->f));
  printf(" x= %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n",
          gsl_vector_get(s->x, 0), gsl_vector_get(s->x, 1),
          gsl_vector_get(s->x, 2), gsl_vector_get(s->x, 3),
          gsl_vector_get(s->x, 4), gsl_vector_get(s->x, 5),
          gsl_vector_get(s->x, 6), gsl_vector_get(s->x, 7),
          gsl_vector_get(s->x, 8), gsl_vector_get(s->x, 9)*dt);
  printf("dx= %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n",
          gsl_vector_get(s->dx, 0), gsl_vector_get(s->dx, 1),
          gsl_vector_get(s->dx, 2), gsl_vector_get(s->dx, 3),
          gsl_vector_get(s->dx, 4), gsl_vector_get(s->dx, 5),
          gsl_vector_get(s->dx, 6), gsl_vector_get(s->dx, 7),
          gsl_vector_get(s->dx, 8), gsl_vector_get(s->dx, 9)*dt);

  return;
}
/*--------------------------------------------------------*/
int modify_header(int *ncards, char ***header)
{
        char  line[CARD_SIZE];
        int   p;

  memset(line, ' ', CARD_SIZE);
  strcpy(line, "HISTORY = 'fit2rotgauss");
  if ((p=add_header_card(line, ncards, header)) == 0)
    printf("\n\tWARNING! modify_header(): add_header_card - failed\n\n");

  return(p);
}
/*--------------------------------------------------------*/
float *write_fit(char *iname, int ncards, char **header,
                int n, double *t, gsl_multifit_fdfsolver *s)
{
        char    *oname;
        int     i,
                ng;
        float   *fit;
        double  A1, B1, C1, D1,
                A2, B2, C2, D2,
                E, S,
                *r, *r1, *r2,
                *g,
                *conv;
        FILE    *outf;

  A1=gsl_vector_get(s->x, 0);
  B1=gsl_vector_get(s->x, 1);
  C1=gsl_vector_get(s->x, 2);
  D1=gsl_vector_get(s->x, 3);
  A2=gsl_vector_get(s->x, 4);
  B2=gsl_vector_get(s->x, 5);
  C2=gsl_vector_get(s->x, 6);
  D2=gsl_vector_get(s->x, 7);
  E=gsl_vector_get(s->x, 8);
  S=fabs(gsl_vector_get(s->x, 9));

  r1=rotprof(n, t, A1, B1, C1, D1);
  r2=rotprof(n, t, A2, B2, C2, D2);

  if ((r=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! write_fit(): calloc(r)\n"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++) r[i]=r1[i]+r2[i];

  free(r1);
  free(r2);

  ng=2*(int)(S*sqrt(-log(1.0e-8)))+1;
  g=gaussprof(ng, S);

  conv=convolve(n, r, ng, g);

  free(r);
  free(g);

  if ((fit=(float *)calloc(n, sizeof(float))) == NULL)
  { perror("\n\tERROR! write_fit: calloc(fit)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++) fit[i]=conv[i]+E;
  free(conv);

  if ((oname=(char *)calloc(strlen(iname)+14, sizeof(char))) == NULL)
  { perror("\n\tERROR! write_fit(): calloc(oname)"); exit(EXIT_FAILURE); }

  strcpy(oname, "fit2rotgauss_");  strcat(oname, iname);
  printf("Output file:\n%s\n", oname);

  if ((outf=fopen(oname, "w")) == NULL) { perror(oname); exit(EXIT_FAILURE); }
  write_FITS_header(outf, ncards, header);
  write_FITS_1Dimage(outf, n, sizeof(float), fit);
  fclose(outf);

  if (needswap()) swap_bytes(fit, n, sizeof(float));

  return(fit);
}
/*--------------------------------------------------------*/
int main(int argc, char *argv[])
{
        char                      *iname,     // input file name
                                  *pname,     // parameters file
                                  hline[CARD_SIZE],
                                  **header;
        int                       i,
                                  ncards,
                                  status,     //
                                  iter;       // number of iterations
        long                      n;          // number of data points
        float                     *fdata,
                                  *fit;
        double                    *t,         // independent variable
                                  *y,         // measured data
                                  *sigma,     // data errors
                                  t0, dt, crpix,
                                  x_init[p];  // initial parameters vector
        gsl_multifit_fdfsolver    *s;
        gsl_matrix                *covar;
        gsl_multifit_function_fdf f;
        gsl_rng                   *r;
        gsl_vector_view           x;

  if (argc != 3)  usage();

  iname=argv[1];
  pname=argv[2];

  read_par(pname);

/* This is the data to be fitted */
  fdata=read_FITS_1Dfile(iname, &ncards, &header, &n);
  if (par.verbose > 0) printf("%s: %ld data points\n", iname, n);

  if (get_FITS_key(ncards, header, "CD1_1", hline) < 1)
  {
    if (get_FITS_key(ncards, header, "CDELT1", hline) < 1)
    {
      printf("\tWARNING! no CD1_1 nor CDELT1 in the FITS header: ");
      printf("using 1.0\n");
      dt=1.0;
    }
    else sscanf(hline, "%lg", &dt);
  }
  else sscanf(hline, "%lg", &dt);

  if (get_FITS_key(ncards, header, "CRVAL1", hline) < 1)
  {
    printf("\tWARNING! CRVAL1 not found in the FITS header: using 0.0\n");
    t0=0.0;
  }
  else sscanf(hline, "%lg", &t0);

  if (get_FITS_key(ncards, header, "CRPIX1", hline) < 1)
  {
    printf("\tWARNING! CRPIX1 not found in the FITS header: using 0.0\n");
    crpix=0.0;
  }
  else sscanf(hline, "%lg", &crpix);

  if (par.verbose > 1) printf("t0= %f   dt= %f   crpix= %f\n", t0, dt, crpix);

  x_init[0]=par.a1;
  x_init[1]=par.b1;
  x_init[2]=par.c1;
  x_init[3]=par.d1;
  x_init[4]=par.a2;
  x_init[5]=par.b2;
  x_init[6]=par.c2;
  x_init[7]=par.d2;
  x_init[8]=par.e;
  x_init[9]=par.s/dt;

  x=gsl_vector_view_array(x_init, p);

  gsl_rng_env_setup();

  type=gsl_rng_default;
  r=gsl_rng_alloc(type);

  if ((y=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fit2rotgauss: calloc(y)"); exit(EXIT_FAILURE); }

  if ((t=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fit2rotgauss: calloc(t)"); exit(EXIT_FAILURE); }

  if ((sigma=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fit2rotgauss: calloc(sigma)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++)
  {
    y[i]=fdata[i];
    t[i]=t0+(i+1-crpix)*dt;
    sigma[i]=1.0;
  }
  free(fdata);

  struct data d = {n, t, y, sigma};

  f.f=&expb_f;
  f.df=&expb_df;
  f.fdf=&expb_fdf;
  f.n=n;
  f.p=p;
  f.params=&d;

  T=gsl_multifit_fdfsolver_lmsder;
  s=gsl_multifit_fdfsolver_alloc(T, n, p);
  gsl_multifit_fdfsolver_set(s, &f, &x.vector);

  iter=0;
  if (par.verbose > 0) print_state(iter, s, dt);

  do
  {
    iter++;
    status=gsl_multifit_fdfsolver_iterate(s);

    if (par.verbose > 0)
    {
      printf("status= %s\n", gsl_strerror (status));
      print_state(iter, s, dt);
    }
    if (status != 0) break;

    status=gsl_multifit_test_delta(s->dx, s->x, par.epsabs, par.epsrel);
  }
  while ((status == GSL_CONTINUE) && (iter < par.max_iter));

  covar=gsl_matrix_alloc(p, p);
  gsl_multifit_covar(s->J, 0.0, covar);

#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

  printf("--------------\n");
  printf(" A1   B1   C1   D1\n");
  printf("%g  %g  %g  %g    +-   %g  %g  %g  %g\n\n",
          FIT(0), FIT(1), FIT(2), FIT(3), ERR(0), ERR(1), ERR(2), ERR(3));

  printf(" A2   B2   C2   D2\n");
  printf("%g  %g  %g  %g    +-   %g  %g  %g  %g\n\n",
          FIT(4), FIT(5), FIT(6), FIT(7), ERR(4), ERR(5), ERR(6), ERR(7));

  printf(" E    S\n");
  printf("%g  %g    +-   %g  %g\n", FIT(8), fabs(FIT(9)*dt), ERR(8), ERR(9)*dt);
  printf("--------------\n");

  double chi=gsl_blas_dnrm2(s->f);
  printf("chisq/dof= %g\n", pow(chi, 2.0)/(n-p));
  printf("status= %s\n", gsl_strerror(status));

  i=modify_header(&ncards, &header);
  fit=write_fit(iname, ncards, header, n, t, s);

  for (i=0; i<ncards; i++) free(header[i]);
  free(header);

  free(fit);

  gsl_multifit_fdfsolver_free(s);

  return(0);
}
/*** END ***/
