/*========================================================*/
/*                                                        */
/*  fitgauss.c          version 0.7.0   2010.01.14        */
/*                                                        */
/*  Copyright (C) 2010 by Wojtek Pych, CAMK PAN           */
/*  pych@camk.edu.pl                                      */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*                                                        */
/*========================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>

#include "pfitsio.h"

struct data
{
  int     n;
  double  *t;
  double  *y;
  double  *sigma;
};

/* Control values */
const int                         p = 4;    // number of fit parameters
const gsl_multifit_fdfsolver_type *T;
const gsl_rng_type                *type;

const int    ITER=500;    // number of "gsl" iterations
const double epsabs=1.0e-11;
const double epsrel=1.0e-11;

void usage(void);
int modify_header(int *, char ***);
float *write_fit(char *, int, char **, int, double *, gsl_multifit_fdfsolver *);
double gaussprof(double, double, double, double);
int expb_f(const gsl_vector *, void *, gsl_vector *);
int expb_df(const gsl_vector *, void *, gsl_matrix *);
int expb_fdf(const gsl_vector *, void *, gsl_vector *, gsl_matrix *);
void print_state(int, gsl_multifit_fdfsolver *);

/*--------------------------------------------------------*/
void usage()
{
  printf("\n\t USAGE:  fitgauss  input_file  A  B  C  D  [-v]\n\n");
  printf("\t Y = A exp[-(t-B)*(t-B)/C/C] + D\n");
  printf("\t -v: verbose\n\n");
  exit(EXIT_FAILURE);
}
/*--------------------------------------------------------*/
int modify_header(int *ncards, char ***header)
{
        char  line[CARD_SIZE];
        int   p;

  memset(line, ' ', CARD_SIZE);
  strcpy(line, "HISTORY = 'fitgauss'");
  if ((p=add_header_card(line, ncards, header)) == 0)
    printf("\n\tWARNING! modify_header(): add_header_card - failed\n\n");

  return(p);
}
/*--------------------------------------------------------*/
float *write_fit(char *iname, int ncards, char **header,
                int n, double *t, gsl_multifit_fdfsolver *s)
{
        char    *oname;
        int     i;
        float   *fit;
        double  A, B, C, D;
        FILE    *outf;

  A=gsl_vector_get(s->x, 0);
  B=gsl_vector_get(s->x, 1);
  C=gsl_vector_get(s->x, 2);
  D=gsl_vector_get(s->x, 3);

  if ((fit=(float *)calloc(n, sizeof(float))) == NULL)
  { perror("\n\tERROR! write_fit(): calloc(fit)"); exit(EXIT_FAILURE); }
  for (i=0; i<n; i++) fit[i]=gaussprof(t[i], A, B, C)+D;

  if ((oname=(char *)calloc(strlen(iname)+10, sizeof(char))) == NULL)
  { perror("\n\tERROR! write_fit(): calloc(oname)"); exit(EXIT_FAILURE); }

  strcpy(oname, "fitgauss_");  strcat(oname, iname);
  printf("Output file:\n%s\n", oname);

  if ((outf=fopen(oname, "w")) == NULL) { perror(oname); exit(EXIT_FAILURE); }
  write_FITS_header(outf, ncards, header);
  write_FITS_1Dimage(outf, n, sizeof(float), fit);
  fclose(outf);

  if (needswap()) swap_bytes(fit, n, sizeof(float));

  return(fit);
}
/*--------------------------------------------------------*/
/*  Model Yi = A * exp[-(t-B)*(t-B)/C/C] + D              */

double gaussprof(double x, double a, double b, double c)
{
        double  y,
                d;

  d=(x-b)/c;
  y=a*exp(-d*d);

  return(y);
}
/*--------------------------------------------------------*/
int expb_f(const gsl_vector *x, void *params, gsl_vector *f)
{
        int  i,
                n;
        double  Yi,
                *t,
                *y,
                *sigma,
                A, B, C, D;

  n=((struct data *)params)->n;
  t=((struct data *)params)->t;
  y=((struct data *)params)->y;
  sigma=((struct data *)params)->sigma;

  A=gsl_vector_get(x, 0);
  B=gsl_vector_get(x, 1);
  C=gsl_vector_get(x, 2);
  D=gsl_vector_get(x, 3);

  for (i=0; i<n; i++)
  {
    Yi=gaussprof(t[i], A, B, C)+D;
    gsl_vector_set(f, i, (Yi - y[i])/sigma[i]);
  }

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
/*  Jacobian matrix J(i,j) = dfi / dxj,                   */
/*  where  fi = (Yi - yi)/sigma[i],                       */
/*         Yi = A * exp(-(t-B)*(t-B)/C/C) + D             */
/*  and the xj are the parameters (A, B, C, D)            */
/*  0:   df/dA (amplitude)                                */
/*  1:   df/dB (shift)                                    */
/*  2:   df/dC (sigma)                                    */
/*  3:   df/dD (zero level)                               */

int expb_df(const gsl_vector *x, void *params, gsl_matrix *J)
{
        int     i,
                n;
        double  w, z,
                *t,
                *sigma,
                A, B, C;

  n=((struct data *)params)->n;
  t=((struct data *)params)->t;
  sigma=((struct data *)params)->sigma;
  A=gsl_vector_get(x, 0);
  B=gsl_vector_get(x, 1);
  C=gsl_vector_get(x, 2);

  for (i=0; i<n; i++)
  {
    w=(t[i]-B)/C;
    z=exp(-w*w)/sigma[i];
    gsl_matrix_set(J, i, 0, z);
    gsl_matrix_set(J, i, 1, 2.0*A/C*w*z);
    gsl_matrix_set(J, i, 2, 2.0*A/C*w*w*z);
    gsl_matrix_set(J, i, 3, 1.0/sigma[i]);

/* Freeze selected parameters */
//    gsl_matrix_set(J, i, 0, 0.0);
//    gsl_matrix_set(J, i, 1, 0.0);
//    gsl_matrix_set(J, i, 2, 0.0);
//    gsl_matrix_set(J, i, 3, 0.0);
  }

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
int expb_fdf(const gsl_vector *x, void *params, gsl_vector *f, gsl_matrix *J)
{
  expb_f(x, params, f);
  expb_df(x, params, J);

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
void print_state(int iter, gsl_multifit_fdfsolver *s)
{
  printf("iter: %3u / %u\n", iter, ITER);
  printf(" x= %g  %g  %g  %g  |f(x)|= %g\n",
          gsl_vector_get(s->x, 0), gsl_vector_get(s->x, 1),
          gsl_vector_get(s->x, 2), gsl_vector_get(s->x, 3),
          gsl_blas_dnrm2(s->f));
  printf("dx= %g  %g  %g  %g\n",
          gsl_vector_get(s->dx, 0), gsl_vector_get(s->dx, 1),
          gsl_vector_get(s->dx, 2), gsl_vector_get(s->dx, 3));

  return;
}
/*--------------------------------------------------------*/
int main(int argc, char *argv[])
{
        short                     verbose;
        int                       i,          // loop numerator
                                  ncards,
                                  iter,       // number of iterations
                                  status;     //
        long                      n;          // number of data points
        char                      *iname,     // input file name
                                  hline[CARD_SIZE],
                                  **header;   // input file name
        float                     *fdata;
        double                    *t,         // independent variable
                                  t0, dt, crpix,
                                  *y,         // measured data
                                  *sigma,     // data errors
                                  chi,
                                  x_init[p];  // initial parameters vector
        gsl_multifit_fdfsolver    *s;
        gsl_matrix                *covar;
        gsl_multifit_function_fdf f;

  verbose=0;
  iname=argv[1];

  if (argc == (p+2))
  {
    for (i=0; i<p; i++) x_init[i]=atof(argv[i+2]);
  }
  else if (argc == (p+3))
  {
    for (i=0; i<p; i++) x_init[i]=atof(argv[i+2]);
    if (tolower(argv[p+2][1])=='v') verbose=1;
  }
  else  usage();

  gsl_vector_view x=gsl_vector_view_array(x_init, p);
  gsl_rng *r;

  gsl_rng_env_setup();

  type=gsl_rng_default;
  r=gsl_rng_alloc(type);

/* This is the data to be fitted */
  fdata=read_FITS_1Dfile(iname, &ncards, &header, &n);
  printf("%s: %ld data points\n", iname, n);

  if (get_FITS_key(ncards, header, "CD1_1", hline) < 1)
  {
    if (get_FITS_key(ncards, header, "CDELT1", hline) < 1)
    {
      printf("\n\tWARNING! no CD1_1 nor CDELT1 found in the FITS header: ");
      printf("using 1.0\n");
      dt=1.0;
    }
    else sscanf(hline, "%lg", &dt);
  }
  else sscanf(hline, "%lg", &dt);

  if (get_FITS_key(ncards, header, "CRVAL1", hline) < 1)
  {
    printf("\n\tWARNING! CRVAL1 not found in the FITS header: using 0.0\n");
    t0=0.0;
  }
  else sscanf(hline, "%lg", &t0);

  if (get_FITS_key(ncards, header, "CRPIX1", hline) < 1)
  {
    printf("\n\tWARNING! CRPIX1 not found in the FITS header: using 0.0\n");
    crpix=0.0;
  }
  else sscanf(hline, "%lg", &crpix);

  if (verbose > 0)
    printf("t0= %15.12f  dt= %15.12f  crpix= %15.12f\n", t0, dt, crpix);

  if ((y=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fitgauss(): calloc(y)"); exit(EXIT_FAILURE); }
  if ((t=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fitgauss(): calloc(t)"); exit(EXIT_FAILURE); }
  if ((sigma=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fitgauss(): calloc(sigma)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++)
  {
    y[i]=fdata[i];
    t[i]=t0+(i+1.0-crpix)*dt;
    sigma[i]=1.0;
  }

  free(fdata);

  struct data d = {n, t, y, sigma};

  f.f=&expb_f;
  f.df=&expb_df;
  f.fdf=&expb_fdf;
  f.n=n;
  f.p=p;
  f.params=&d;

  T=gsl_multifit_fdfsolver_lmsder;
  s=gsl_multifit_fdfsolver_alloc(T, n, p);
  gsl_multifit_fdfsolver_set(s, &f, &x.vector);

  iter=0;
  if (verbose > 0) print_state(iter, s);
  do
  {
    iter++;
    status=gsl_multifit_fdfsolver_iterate(s);

    if (verbose > 0)
    {
      printf("status= %s\n", gsl_strerror (status));
      print_state(iter, s);
    }

    if (status != 0) break;

    status=gsl_multifit_test_delta(s->dx, s->x, epsabs, epsrel);
  }
  while ((status == GSL_CONTINUE) && (iter < ITER));

  covar=gsl_matrix_alloc(p, p);
  gsl_multifit_covar(s->J, 0.0, covar);

#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

  if (verbose > 0)
  {
    printf(" A= %12.8f  B= %12.8f  C= %12.8f  D= %12.8f\n",
             FIT(0), FIT(1), FIT(2), FIT(3));
    printf("dA= %12.8f dB= %12.8f dC= %12.8f dD= %12.8f\n",
            ERR(0), ERR(1), ERR(2), ERR(3));
  }

  chi=gsl_blas_dnrm2(s->f);

  printf("chisq/dof= %g\n", pow(chi, 2.0)/(n-p));
  printf ("status= %s\n", gsl_strerror(status));

#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

  printf(" A= %12.8f  B= %12.8f  C= %12.8f  D= %12.8f\n",
          FIT(0), FIT(1), FIT(2), FIT(3));
  printf("dA= %12.8f dB= %12.8f dC= %12.8f dD= %12.8f\n",
          ERR(0), ERR(1), ERR(2), ERR(3));

  ncards=modify_header(&ncards, &header);
  fdata=write_fit(iname, ncards, header, n, t, s);

  free(fdata);
  gsl_multifit_fdfsolver_free(s);

  return(0);
}
/*** END ***/
