/*========================================================*/
/*                                                        */
/*  fitrotgauss.c       version 1.3.0     2010.01.15      */
/*                                                        */
/*  Copyright (C) 2010 by Wojtek Pych, CAMK PAN           */
/*  pych@camk.edu.pl                                      */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*                                                        */
/*  Make a nonlinear least squares fit of                 */
/*  a rotational profile convolved with a gaussian.       */
/*                                                        */
/*  The rotational profile defined as:                    */
/*                                                        */
/*  y = A [ (1 - B) sqrt(z) + Pi/4 B z ] + E              */
/*  where z = (1 - a*a); a = (v - C)/D                    */
/*  A: amplitude                                          */
/*  B: linear limb darkening coefficient                  */
/*  C: radial velocity of the center of mass              */
/*  D: Vsini (rotation speed)                             */
/*                                                        */
/*  Gaussian defined as:                                  */
/*                                                        */
/*  y = 1/sqrt(Pi)/S * exp[ -(i-n/2)*(i-n/2)/S/S ]        */
/*                                                        */
/*========================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>

#include "pfitsio.h"

struct data
{
  int     n;
  double  *t,
          *y,
          *sigma;
};

typedef struct
{
  short   va, vb, vc, vd,
          ve, vs,
          verbose;
  int     max_iter;
  double  a, b, c, d, e, s,
          epsabs,
          epsrel;
} PARAMS;

        PARAMS                    par;

/* Control values */
const size_t                      p = 6;   // number of fit parameters
const gsl_multifit_fdfsolver_type *T;
const gsl_rng_type                *type;

#define STRLEN 200

void    usage(void);
void    read_par(char *);
double  *rotprof(int, double *, double, double, double, double);
double  *gaussprof(int, double);
double  *convolve(int, double *, int, double *);
int     expb_f(const gsl_vector *, void *, gsl_vector *);
int     expb_df(const gsl_vector *, void *, gsl_matrix *);
int     expb_fdf(const gsl_vector *, void *, gsl_vector *, gsl_matrix *);
int     modify_header(int *, char ***);
void    print_state(int, gsl_multifit_fdfsolver *, double);
void 	write_fit(char *, int, char **, int, double *, gsl_multifit_fdfsolver *);

/*--------------------------------------------------------*/
void usage()
{
  printf("\n\t USAGE:  fitrotgauss  input_file  parameters_file\n\n");
  exit(EXIT_FAILURE);
}
/*--------------------------------------------------------*/
void read_par(char *parfname)
{
          char  line[STRLEN],
                key[STRLEN],
                val[STRLEN];
          FILE  *inpf;

  if ((inpf = fopen(parfname, "r")) == NULL)
  { perror(parfname); exit(EXIT_FAILURE); }

  while (feof(inpf) == 0)
  {
    if (fgets(line, STRLEN, inpf) == NULL)
    { perror("\n\tERROR! read_par()\n"); exit(EXIT_FAILURE); }
    if (feof(inpf)) break;

    sscanf(line, "%s = %s", key, val);

    if (strcasecmp(key, "END") == 0) break;
    else if (strcasecmp(key, "A") == 0)       sscanf(val, "%lf", &par.a);
    else if (strcasecmp(key, "B") == 0)       sscanf(val, "%lf", &par.b);
    else if (strcasecmp(key, "C") == 0)       sscanf(val, "%lf", &par.c);
    else if (strcasecmp(key, "D") == 0)       sscanf(val, "%lf", &par.d);
    else if (strcasecmp(key, "E") == 0)       sscanf(val, "%lf", &par.e);
    else if (strcasecmp(key, "S") == 0)       sscanf(val, "%lf", &par.s);
    else if (strcasecmp(key, "VAR_A") == 0)   sscanf(val, "%hd", &par.va);
    else if (strcasecmp(key, "VAR_B") == 0)   sscanf(val, "%hd", &par.vb);
    else if (strcasecmp(key, "VAR_C") == 0)   sscanf(val, "%hd", &par.vc);
    else if (strcasecmp(key, "VAR_D") == 0)   sscanf(val, "%hd", &par.vd);
    else if (strcasecmp(key, "VAR_E") == 0)   sscanf(val, "%hd", &par.ve);
    else if (strcasecmp(key, "VAR_S") == 0)   sscanf(val, "%hd", &par.vs);
    else if (strcasecmp(key, "EPSABS") == 0)  sscanf(val, "%lf", &par.epsabs);
    else if (strcasecmp(key, "EPSREL") == 0)  sscanf(val, "%lf", &par.epsrel);
    else if (strcasecmp(key, "MAXITER") == 0) sscanf(val, "%d",  &par.max_iter);
    else if (strcasecmp(key, "VERBOSE") == 0) sscanf(val, "%hd", &par.verbose);
    else printf("\n\tWARNING: Unknown parameter %s\n", key);
  }

  fclose(inpf);

  if (par.verbose > 0)
  {
    printf("Reading: %s\n", parfname);
    printf("-----------\n");
    printf("A       = %g\n", par.a);
    printf("B       = %g\n", par.b);
    printf("C       = %g\n", par.c);
    printf("D       = %g\n", par.d);
    printf("E       = %g\n", par.e);
    printf("S       = %g\n", par.s);
    printf("VAR_A   = %d\n", par.va);
    printf("VAR_B   = %d\n", par.vb);
    printf("VAR_C   = %d\n", par.vc);
    printf("VAR_D   = %d\n", par.vd);
    printf("VAR_E   = %d\n", par.ve);
    printf("VAR_S   = %d\n", par.vs);
    printf("EPSABS  = %g\n", par.epsabs);
    printf("EPSREL  = %g\n", par.epsrel);
    printf("VERBOSE = %d\n", par.verbose);
    printf("-----------\n");
  }
  
  return;
}
/*--------------------------------------------------------*/
double *rotprof(int n, double *t, double a, double b, double c, double d)
{
        int     i;
        double  *y,
                z;

  if ((y=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! rotprof: calloc(y)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++)
  {
    z=(t[i]-c)/d;
    z=1.0-z*z;
    if (z > 0.0) y[i]=a*((1.0-b)*sqrt(z)+b*M_PI/4.0*z);
    else         y[i]=0.0;
  }

  return(y);
}
/*--------------------------------------------------------*/
double *gaussprof(int n, double s)
{
        int     i,
                k;
        double  *y,
                *g,
                a,    // normalization
                b;

  if ((y=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! gaussprof: calloc(y)"); exit(EXIT_FAILURE); }

  k=n/2;
  g=y+k;

  a=sqrt(M_PI)*s;
  for (i=-k; i<=k; i++)
  {
    b=i/s;
    g[i]=exp(-b*b)/a;
  }

  return(y);
}
/*--------------------------------------------------------*/
double *convolve(int nr, double *r, int ng, double *g)
{
        int     i, j,
                k;
        double  *cg,
                *lconv;

  k=ng/2;
  if ((lconv=(double *)calloc(nr, sizeof(double))) == NULL)
  { perror("\n\tERROR! calloc(conv)"); exit(EXIT_FAILURE); }

  cg=g+k;

  for (i=0; i<nr; i++)
  {
    lconv[i]=0.0;
    for (j=-k; j<=k; j++)
    {
      if ((i+j >= 0) && (i+j < nr))  lconv[i]+=r[i+j]*cg[j];
    }
  }

  return(lconv);
}
/*--------------------------------------------------------*/
int expb_f(const gsl_vector *x, void *params, gsl_vector *f)
{
        int     i,
                n,
                ng;
        double  A, B, C, D,
                E, S,
                *t,
                *y,
                *sigma,
                *r,
                *g,
                *conv;

  n=((struct data *)params)->n;
  t=((struct data *)params)->t;
  y=((struct data *)params)->y;
  sigma=((struct data *)params)->sigma;

  A=gsl_vector_get(x, 0);
  B=gsl_vector_get(x, 1);
  C=gsl_vector_get(x, 2);
  D=gsl_vector_get(x, 3);
  E=gsl_vector_get(x, 4);
  S=gsl_vector_get(x, 5);

  r=rotprof(n, t, A, B, C, D);

  S=fabs(S);
  ng=2*(int)(S*sqrt(-log(1.0e-8)))+1;
  if (par.verbose > 1) printf("gaussprof: S= %f   ng= %d\n", S, ng);
  g=gaussprof(ng, S);

  conv=convolve(n, r, ng, g);

  free(r);
  free(g);

  for (i=0; i<n; i++)
    gsl_vector_set(f, i, (conv[i] + E - y[i])/sigma[i]);

  free(conv);

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
/* Jacobian matrix J(i,j) = dfi / dxj,                              */
/* where  fi = (Yi - yi)/sigma[i],                                  */
/*        z=1-(t-C)*(t-C)/D/D                                       */
/*        Yi = A * [(1-B)*sqrt(z)+B*M_PI/4.0*z] + E                 */
/* and the xj are the parameters (A, B, C, D, E)                    */
/*                                                                  */
/* 0:   df/dA (amplitude)           */
/* 1:   df/dB (limb darkening)      */
/* 2:   df/dC (radial velocity)     */
/* 3:   df/dD (Vsini)               */
/* 4:   df/dE (zero level)          */
/* 5:   df/dS (Gaussian sigma)      */

int expb_df(const gsl_vector *x, void *params, gsl_matrix *J)
{
        int     i,
                n,
                ng;
        double  v, u, z,
                *t,
                *sigma,
                A, B, C, D,
                E, S,
                *g,
                *r,
                *conv,
                *dfa, *dfb, *dfc, *dfd,
                *dfe, *dfs, *ldfs;

  n=((struct data *)params)->n;
  t=((struct data *) params)->t;
  sigma=((struct data *) params)->sigma;

  A=gsl_vector_get(x, 0);
  B=gsl_vector_get(x, 1);
  C=gsl_vector_get(x, 2);
  D=gsl_vector_get(x, 3);
  E=gsl_vector_get(x, 4);
  S=gsl_vector_get(x, 5);

  if ((dfa=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df: calloc(dfa)"); exit(EXIT_FAILURE); }

  if ((dfb=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df: calloc(dfb)"); exit(EXIT_FAILURE); }

  if ((dfc=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df: calloc(dfc)"); exit(EXIT_FAILURE); }

  if ((dfd=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df: calloc(dfd)"); exit(EXIT_FAILURE); }

  if ((dfe=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df: calloc(dfe)"); exit(EXIT_FAILURE); }

  if ((dfs=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! expb_df: calloc(dfs)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++)
  {
    v=(t[i]-C)/D;
    z=1.0-v*v;

    if (z > 0.0)
    {
      if (par.va) dfa[i]=((1.0-B)*sqrt(z) + B*M_PI/4.0*z)/sigma[i];
      else        dfa[i]=0.0;

      if (par.vb) dfb[i]=A*(M_PI/4.0*z - sqrt(z))/sigma[i];
      else        dfb[i]=0.0;

      u=((1.0-B)/sqrt(z) + M_PI/2.0*B)/sigma[i];

      if (par.vc) dfc[i]=A*v/D*u;
      else        dfc[i]=0.0;

      if (par.vd) dfd[i]=A*v*v/D*u;
      else        dfd[i]=0.0;
    }
    else
    {
      dfa[i]=0.0;
      dfb[i]=0.0;
      dfc[i]=0.0;
      dfd[i]=0.0;
    }

    if (par.ve) dfe[i]=1.0;
    else        dfe[i]=0.0;
  }

  S=fabs(S);
  ng=2*(int)(S*sqrt(-log(1.0e-8)))+1;
  if (par.verbose > 2) printf("expb_df: S= %G  ng= %d\n", S, ng);
  g=gaussprof(ng, S);

  if (par.va)
  {
    conv=convolve(n, dfa, ng, g);
    for (i=0; i<n; i++) dfa[i]=conv[i];
    free(conv);
  }

  if (par.vb)
  {
    conv=convolve(n, dfb, ng, g);
    for (i=0; i<n; i++) dfb[i]=conv[i];
    free(conv);
  }

  if (par.vc)
  {
    conv=convolve(n, dfc, ng, g);
    for (i=0; i<n; i++) dfc[i]=conv[i];
    free(conv);
  }

  if (par.vd)
  {
    conv=convolve(n, dfd, ng, g);
    for (i=0; i<n; i++) dfd[i]=conv[i];
    free(conv);
  }

  free(g);

  if (par.vs)
  {
    r=rotprof(n, t, A, B, C, D);

    if ((ldfs=(double *)calloc(ng, sizeof(double))) == NULL)
    { perror("\n\tERROR! expb_df: calloc(ldfs)"); exit(EXIT_FAILURE); }

    for (i=0; i<ng; i++)
    {
      z=(i-ng/2)/S;
      ldfs[i]=2.0*z*z-1.0;
      ldfs[i]*=exp(-z*z)/sqrt(M_PI)/S/S;
    }

    conv=convolve(n, r, ng, ldfs);

    for (i=0; i<n; i++) dfs[i]=conv[i];

    free(ldfs);
    free(conv);
    free(r);
  }
  else
    for (i=0; i<n; i++) dfs[i]=0.0;

  for (i=0; i<n; i++)
  {
    gsl_matrix_set(J, i, 0, dfa[i]);
    gsl_matrix_set(J, i, 1, dfb[i]);
    gsl_matrix_set(J, i, 2, dfc[i]);
    gsl_matrix_set(J, i, 3, dfd[i]);
    gsl_matrix_set(J, i, 4, dfe[i]);
    gsl_matrix_set(J, i, 5, dfs[i]);
  }

  free(dfa);
  free(dfb);
  free(dfc);
  free(dfd);
  free(dfe);
  free(dfs);

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
int expb_fdf(const gsl_vector *x, void *params, gsl_vector *f, gsl_matrix *J)
{
  expb_f(x, params, f);
  expb_df(x, params, J);

  return GSL_SUCCESS;
}
/*--------------------------------------------------------*/
void print_state(int iter, gsl_multifit_fdfsolver *s, double dt)
{
  printf("iter: %3d    |f(x)|= %g\n", iter, gsl_blas_dnrm2(s->f));
  printf(" x= %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n",
          gsl_vector_get(s->x, 0), gsl_vector_get(s->x, 1),
          gsl_vector_get(s->x, 2), gsl_vector_get(s->x, 3),
          gsl_vector_get(s->x, 4), gsl_vector_get(s->x, 5)*dt);
  printf("dx= %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n",
          gsl_vector_get(s->dx, 0), gsl_vector_get(s->dx, 1),
          gsl_vector_get(s->dx, 2), gsl_vector_get(s->dx, 3),
          gsl_vector_get(s->dx, 4), gsl_vector_get(s->dx, 5)*dt);

  return;
}
/*--------------------------------------------------------*/
int modify_header(int *ncards, char ***header)
{
        char  line[CARD_SIZE];
        int   p;

  memset(line, ' ', CARD_SIZE);
  strcpy(line, "HISTORY = 'fitrotgauss");

  if ((p=add_header_card(line, ncards, header)) == 0)
    printf("\n\tWARNING! modify_header(): add_header_card - failed\n\n");

  return(p);
}
/*--------------------------------------------------------*/
void write_fit(char *iname, int ncards, char **header, int n, double *t,
  gsl_multifit_fdfsolver *s)
{
        char    *oname;
        int     i,
                ng;
        float   *tmp;
        double  A, B, C, D,
                E, S,
                *r,
                *g,
                *conv;
        FILE    *outf;

  A=gsl_vector_get(s->x, 0);
  B=gsl_vector_get(s->x, 1);
  C=gsl_vector_get(s->x, 2);
  D=gsl_vector_get(s->x, 3);
  E=gsl_vector_get(s->x, 4);
  S=gsl_vector_get(s->x, 5);

  r=rotprof(n, t, A, B, C, D);

  ng=2*(int)(S*sqrt(-log(1.0e-8)))+1;
  g=gaussprof(ng, S);

  conv=convolve(n, r, ng, g);

  free(r);
  free(g);

  if ((tmp=(float *)calloc(n, sizeof(float))) == NULL)
  { perror("\n\tERROR! write_fit: calloc(tmp)"); exit(EXIT_FAILURE); }
  for (i=0; i<n; i++) tmp[i]=conv[i]+E;
  free(conv);

  if ((oname=(char *)calloc(strlen(iname)+14, sizeof(char))) == NULL)
  { perror("\n\tERROR! write_fit: calloc(oname)"); exit(EXIT_FAILURE); }

  strcpy(oname, "fitrotgauss_");  strcat(oname, iname);
  printf("Output file: %s\n", oname);

  if ((outf=fopen(oname, "w")) == NULL) { perror(oname); exit(EXIT_FAILURE); }
  write_FITS_header(outf, ncards, header);
  write_FITS_1Dimage(outf, n, sizeof(float), tmp);
  fclose(outf);

  free(tmp);

  return;
}
/*--------------------------------------------------------*/
int main(int argc, char *argv[])
{
        int                       i,
                                  ncards,
                                  status,     //
                                  iter;       // number of iterations
        long                      n;          // number of data points
        char                      *iname,     // input file name
                                  *pname,     // parameters file
                                  hline[CARD_SIZE],
                                  **header;
        float                     *fdata;
        double                    *t,         // independent variable
                                  *y,         // measured data
                                  *sigma,     // data errors
                                  t0, dt, crpix,
                                  x_init[p];  // initial parameters vector
        gsl_multifit_fdfsolver    *s;
        gsl_matrix                *covar;
        gsl_multifit_function_fdf f;
        gsl_rng                   *r;
        gsl_vector_view           x;

  if (argc != 3)  usage();

  iname=argv[1];
  pname=argv[2];

  read_par(pname);

/* This is the data to be fitted */
  fdata=read_FITS_1Dfile(iname, &ncards, &header, &n);
  if (par.verbose > 0) printf("%s: %ld data points\n", iname, n);

  if (get_FITS_key(ncards, header, "CD1_1", hline) < 1)
  {
    if (get_FITS_key(ncards, header, "CDELT1", hline) < 1)
    {
      printf("\n\tWARNING! no CD1_1 nor CDELT1 in the FITS header: ");
      printf("using 1.0\n");
      dt=1.0;
    }
    else sscanf(hline, "%lg", &dt);
  }
  else sscanf(hline, "%lg", &dt);

  if (get_FITS_key(ncards, header, "CRVAL1", hline) < 1)
  {
    printf("\n\tWARNING! CRVAL1 not found in the FITS header: using 0.0\n");
    t0=0.0;
  }
  else sscanf(hline, "%lg", &t0);

  if (get_FITS_key(ncards, header, "CRPIX1", hline) < 1)
  {
    printf("\n\tWARNING! CRPIX1 not found in the FITS header: using 0.0\n");
    crpix=0.0;
  }
  else sscanf(hline, "%lg", &crpix);

  if (par.verbose > 1) printf("t0= %f   dt= %f   crpix= %f\n", t0, dt, crpix);

  x_init[0]=par.a;
  x_init[1]=par.b;
  x_init[2]=par.c;
  x_init[3]=par.d;
  x_init[4]=par.e;
  x_init[5]=par.s/dt;

  x=gsl_vector_view_array(x_init, p);

  gsl_rng_env_setup();

  type=gsl_rng_default;
  r=gsl_rng_alloc(type);

  if ((y=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fitrotgauss(): calloc(y)"); exit(EXIT_FAILURE); }
  if ((t=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fitrotgauss(): calloc(t)"); exit(EXIT_FAILURE); }
  if ((sigma=(double *)calloc(n, sizeof(double))) == NULL)
  { perror("\n\tERROR! fitrotgauss(): calloc(sigma)"); exit(EXIT_FAILURE); }

  for (i=0; i<n; i++)
  {
    y[i]=fdata[i];
    t[i]=t0+(i+1-crpix)*dt;
    sigma[i]=1.0;
  }
  free(fdata);

  struct data d = {n, t, y, sigma};

  if (par.verbose > 1) printf("t[0]= %f   t[n-1]= %f\n", t[0], t[n-1]);

  f.f=&expb_f;
  f.df=&expb_df;
  f.fdf=&expb_fdf;
  f.n=n;
  f.p=p;
  f.params=&d;

  T=gsl_multifit_fdfsolver_lmsder;
  s=gsl_multifit_fdfsolver_alloc(T, n, p);
  gsl_multifit_fdfsolver_set(s, &f, &x.vector);

  iter=0;
  if (par.verbose> 0) print_state(iter, s, dt);

  do
  {
    iter++;
    status=gsl_multifit_fdfsolver_iterate(s);

    if (par.verbose > 0)
    {
      printf("status= %s\n", gsl_strerror (status));
      print_state(iter, s, dt);
    }
    if (status != 0) break;

    status=gsl_multifit_test_delta(s->dx, s->x, par.epsabs, par.epsrel);
  }
  while ((status == GSL_CONTINUE) && (iter < par.max_iter));

  covar=gsl_matrix_alloc(p, p);
  gsl_multifit_covar(s->J, 0.0, covar);

#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

  printf("--------------\n");
  printf("A  B  C  D  E  S\n");
  printf("%g  %g  %g  %g  %g  %g\n",
          FIT(0), FIT(1), FIT(2), fabs(FIT(3)), FIT(4), FIT(5)*dt);
  printf("dA  dB  dC  dD  dE  dS\n");
  printf("%g  %g  %g  %g  %g  %g\n",
          ERR(0), ERR(1), ERR(2), ERR(3), ERR(4), ERR(5)*dt);
  printf("--------------\n");

  double chi=gsl_blas_dnrm2(s->f);
  printf("chisq/dof= %g\n", pow(chi, 2.0)/(n-p));
  printf("status= %s\n", gsl_strerror(status));

  i=modify_header(&ncards, &header);
  write_fit(iname, ncards, header, n, t, s);

  gsl_multifit_fdfsolver_free(s);

  for (i=0; i<ncards; i++) free(header[i]);
  free(header);

  return(0);
}
/*** END ***/
