/*========================================================*/
/*                                                        */
/*  pfitsio.c       version 3.2.0         2010.01.14      */
/*                                                        */
/*  Copyright (C) 2008-2010 by Wojtek Pych, CAMK PAN      */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*  Private FITS I/O library.                             */
/*                                                        */
/*  Definition of the FITS:                               */
/*  Hanisch, R. J.; et al. 2001, A&A, 376, 359-380        */
/*                                                        */
/*========================================================*/

#include <math.h>

#include "pfitsio.h"

/*--------------------------------------------------------*/
double scale(const long size, float *data, short *idata, double *fitsbzero)
{
        long    i;
        float   min,
                max;
        double  FITS_bscale,
                FITS_bzero;

/** scalling factors **/
  min = max = data[0];
  for (i=1; i<size; i++)
  {
    if (data[i] > max) max=data[i];
    if (data[i] < min) min=data[i];
  }
  FITS_bzero=(max-min)/2.0;
  FITS_bscale=(max-min)/65535.0;

  for (i=0; i<size; i++)
    idata[i]=(short)rint((data[i]-FITS_bzero)/FITS_bscale);

  *fitsbzero=FITS_bzero;

  return(FITS_bscale);
}
/*--------------------------------------------------------*/
void write_FITS_1Dimage(FILE *outf, const long nx, const int bytepix,
  void *data)
{
        char  *tmp;
        long  size,   // image size in bytes
              ls;

  if (needswap() != 0) swap_bytes(data, nx, bytepix);

  if (fwrite(data, bytepix, nx, outf) != nx)
  {
    perror("\n\tERROR! write_FITS_1Dimage(): fwrite(data)");
    exit(EXIT_FAILURE);
  }

  size=nx*bytepix;
  if ((ls=size%RECORD_SIZE) != 0) ls=RECORD_SIZE-ls;
  if ((tmp=(char *)malloc(ls)) == NULL)
  {
    perror("\n\tERROR! write_FITS_1Dimage(): malloc(tmp)"); exit(EXIT_FAILURE);
  }

  memset(tmp, 0, ls);
  if (fwrite(tmp, 1, ls, outf) != ls)
  {
    printf("\n\tERROR! write_FITS_1Dimage(): writing zero padding\n");
    exit(EXIT_FAILURE);
  }

  free(tmp);

  return;
}
/*--------------------------------------------------------*/
void write_FITS_2Dimage(FILE *outf, const long nx, const long ny,
  const int bytepix, void **data)
{
        char  *tmp;
        int   i;
        long  size,   // image size in bytes
              ls;

  if (needswap() != 0)
  {
    for (i=0; i<ny; i++)
    {
      swap_bytes(data[i], nx, bytepix);
      if (fwrite(data[i], bytepix, nx, outf) != nx)
      {
        perror("\n\tERROR! write_FITS_2Dimage(): fwrite(data[i])");
        exit(EXIT_FAILURE);
      }
    }
  }
  else
  {
    for (i=0; i<ny; i++)
    {
      if (fwrite(data[i], bytepix, nx, outf) != nx)
      {
        perror("\n\tERROR! write_FITS_2Dimage(): fwrite(data[i])");
        exit(EXIT_FAILURE);
      }
    }
  }

  size=nx*ny*bytepix;
  if ((ls=size%RECORD_SIZE) != 0) ls=RECORD_SIZE-ls;
  if ((tmp=(char *)malloc(ls)) == NULL)
  {
    perror("\n\tERROR! write_FITS_2Dimage(): malloc(tmp)"); exit(EXIT_FAILURE);
  }

  memset(tmp, 0, ls);
  if (fwrite(tmp, 1, ls, outf) != ls)
  {
    printf("\n\tERROR! write_FITS_2Dimage(): writing zero padding\n");
    exit(EXIT_FAILURE);
  }

  free(tmp);

  return;
}
/*--------------------------------------------------------*/
void write_FITS_1Dfile(char *oname, const int ncards, char **header,
  const long nx, const int bytepix, void *data)
{
        FILE  *outf;

  if ((outf=fopen(oname, "w")) == NULL)
  {
    printf("\n\tERROR! write_FITS_1Dfile(): "); perror(oname);
    exit(EXIT_FAILURE);
  }

  write_FITS_header(outf, ncards, header);
  write_FITS_1Dimage(outf, nx, bytepix, data);

  if (fclose(outf) != 0)
    printf("\n\tERROR! write_FITS_1Dfile(): cannot close %s\n", oname);

  return;
}
/*--------------------------------------------------------*/
void write_FITS_2Dfile(char *oname, const int ncards, char **header,
                        const long nx, const long ny, const int bytepix,
                        void **data)
{
        FILE  *outf;

  if ((outf=fopen(oname, "w")) == NULL)
  {
    printf("\n\tERROR! write_FITS_2Dfile(): "); perror(oname);
    exit(EXIT_FAILURE);
  }

  write_FITS_header(outf, ncards, header);
  write_FITS_2Dimage(outf, nx, ny, bytepix, data);

  if (fclose(outf) != 0)
    printf("\n\tERROR! write_FITS_2Dfile(): cannot close %s\n", oname);

  return;
}
/*--------------------------------------------------------*/
/*** END ***/
