/*========================================================*/
/*                                                        */
/*  bf.c            2010.01.13      version 1.2.1         */
/*                                                        */
/*  Copyright (C) 2010 by Wojtek Pych, CAMK PAN           */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*                                                        */
/*  Calculate Broadening Function.                        */
/*  Analog of BFpro2 by S. Rucinski in IDL.               */ 
/*                                                        */
/*========================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "pfitsio.h"

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

#include <gsl/gsl_linalg.h>

#define STDSTRLEN 200

void    usage(void);

/*--------------------------------------------------------*/
void usage()
{
  printf("\n\t USAGE:  bf  template_svd  spectrum  [-v]\n\n");
  exit(EXIT_FAILURE);
}
/*--------------------------------------------------------*/
double *dread_FITS_1Dimage(FILE *inf, int hsize, char **header, int *n)
{
        void    *buf;
        char    hline[CARD_SIZE];
        int     i,
                bitpix,
                naxes,
                nx;
        double  *ldata;

  read_FITS_image(inf, header, hsize, &buf);

  if (get_FITS_key(hsize, header, "NAXIS", hline) < 1)
  {
    printf("\n\tERROR! dread_FITS_1Dimage: NAXIS not found in header\n");
    exit(EXIT_FAILURE);
  }
  sscanf(hline, "%d", &naxes);
  if (naxes != 1)
  {
    printf("\n\tERROR! dread_FITS_1Dimage: Input image must be 1 dimensional");
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hsize, header, "NAXIS1", hline) < 1)
  {
    printf("\n\tERROR! dread_FITS_1Dimage: NAXIS1 not found in header\n");
    exit(EXIT_FAILURE);
  }
  sscanf(hline, "%d", &nx);

  if (get_FITS_key(hsize, header, "BITPIX", hline) < 1)
  {
    printf("\n\tERROR! dread_FITS_1Dimage: BITPIX not found in header\n");
    exit(EXIT_FAILURE);
  }
  sscanf(hline, "%d", &bitpix);

  if (bitpix != -64)
  {
    printf("\n\tdread_FITS_1Dimage: BITPIX = %d (should be -64)\n", bitpix);
    printf("\tInput FITS file is should contain double precision data.\n");
    printf("\tExiting!\n\n");
    exit(EXIT_FAILURE);
  }

  if ((ldata=(double *)calloc(nx, sizeof(double))) == NULL)
  {
    perror("\n\tERROR! dread_FITS_1Dimage: calloc(ldata)"); exit(EXIT_FAILURE);
  }
  for (i=0; i<nx; i++) ldata[i]=((double *)buf)[i];

  free(buf);

  *n=nx;

  return(ldata);
}
/*--------------------------------------------------------*/
double *dread_FITS_2Dimage(FILE *inf, int hsize, char **header,
                            int *n1, int *n2)
{
        void    *buf;
        char    hline[CARD_SIZE];
        int     i,
                bitpix,
                naxes,
                nx,
                ny;
        double  *ldata;

  read_FITS_image(inf, header, hsize, &buf);

  if (get_FITS_key(hsize, header, "NAXIS", hline) < 1)
  {
    printf("\n\tERROR! dread_FITS_2Dimage: NAXIS not found in header\n");
    exit(EXIT_FAILURE);
  }
  sscanf(hline, "%d", &naxes);
  if (naxes != 2)
  {
    printf("\n\tERROR! dread_FITS_2Dimage: Input image must be 2 dimensional");
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hsize, header, "NAXIS1", hline) < 1)
  {
    printf("\n\tERROR! dread_FITS_2Dimage: NAXIS1 not found in header\n");
    exit(EXIT_FAILURE);
  }
  sscanf(hline, "%d", &nx);

  if (get_FITS_key(hsize, header, "NAXIS2", hline) < 1)
  {
    printf("\n\tERROR! dread_FITS_2Dimage: NAXIS2 not found in header\n");
    exit(EXIT_FAILURE);
  }
  sscanf(hline, "%d", &ny);

  if (get_FITS_key(hsize, header, "BITPIX", hline) < 1)
  {
    printf("\n\tERROR! dread_FITS_2Dimage: BITPIX not found in header\n");
    exit(EXIT_FAILURE);
  }
  sscanf(hline, "%d", &bitpix);

  if (bitpix != -64)
  {
    printf("\n\tdread_FITS_2Dfile: BITPIX = %d (should be -64)\n", bitpix);
    printf("\tInput FITS file is should contain double precision data.\n");
    printf("\tExiting!\n\n");
    exit(EXIT_FAILURE);
  }

  if ((ldata=(double *)calloc(nx*ny, sizeof(double))) == NULL)
  {
    perror("\n\tERROR! dread_FITS_2Dimage: calloc(ldata)"); exit(EXIT_FAILURE);
  }

  for (i=0; i<nx*ny; i++) ldata[i]=((double *)buf)[i];

  free(buf);

  *n1=nx;
  *n2=ny;

  return(ldata);
}
/*--------------------------------------------------------*/
double read_svd(char vopt, char *iname, int *m, int *n,
               gsl_vector **S, gsl_matrix **U, gsl_matrix **V, double *crpix)
{
  const double C=2.99792458e5;

        char        ***header,
                    hline[CARD_SIZE];
        int         *hsize,
                    i, j,
                    ne,           // number of extensions in FITS file
                    ln, lm;
        long        *offset;
        double      dv,
                    lcrpix,
                    *buf;
        FILE        *inf;
        gsl_vector  *lS;
        gsl_matrix  *lU, *lV;

  if ((inf=fopen(iname, "r")) == NULL) { perror(iname); exit(EXIT_FAILURE); }
  ne=read_FITS_headers(inf, &hsize, &header, &offset);
  if (ne != 2)
  {
    printf("\n\tERROR! Number of extensions in SVD file is %d instead of 2\n",
      ne);
    exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hsize[0], header[0], "O_CD1_1", hline) < 1)
  {
    if (get_FITS_key(hsize[0], header[0], "O_CDELT1", hline) < 1)
    {
      printf("\n\tERROR! read_svd(): no O_CD1_1 nor O_CDELT1 in FITS header\n");
      exit(EXIT_FAILURE);
    }
  }

  sscanf(hline, "%lf", &dv);
  dv*=C*log(10.0);

  if (get_FITS_key(hsize[0], header[0], "O_CRPIX1", hline) < 1)
  {
      printf("\n\tERROR! read_svd(): no O_CRPIX1 in FITS header\n");
      exit(EXIT_FAILURE);
  }
  sscanf(hline, "%lf", &lcrpix);

/* Read vector S */
  fseek(inf, offset[0]+hsize[0]*CARD_SIZE, SEEK_SET);
  buf=dread_FITS_1Dimage(inf, hsize[0], header[0], &lm);
  lS=gsl_vector_calloc(lm);
  memcpy(lS->data, buf, lm*sizeof(double));
  free(buf);
  *m=lm;

/* Read matrix U */
  fseek(inf, offset[1]+hsize[1]*CARD_SIZE, SEEK_SET);
  buf=dread_FITS_2Dimage(inf, hsize[1], header[1], &ln, &lm);
  lU=gsl_matrix_calloc(ln, lm);
  memcpy(lU->data, buf, lm*ln*sizeof(double));
  free(buf);
  if (*m != lm)
  { printf("\n\tERROR! read_svd: m != m (U)\n"); exit(EXIT_FAILURE); }
  *n=ln;

/* Read matrix VT */
  fseek(inf, offset[2]+hsize[2]*CARD_SIZE, SEEK_SET);
  buf=dread_FITS_2Dimage(inf, hsize[2], header[2], &lm, &lm);
  lV=gsl_matrix_calloc(lm, lm);
  memcpy(lV->data, buf, lm*lm*sizeof(double));
  free(buf);
  if (*m != lm)
  { printf("\n\tERROR! read_svd: m != m (V)\n"); exit(EXIT_FAILURE); }

  fclose(inf);

  for (i=0; i<ne; i++)
  {
    for (j=0; j<hsize[i]; j++) free(header[i][j]);
    free(header[i]);
  }
  free(header);

  *S=lS;
  *U=lU;
  *V=lV;

  *crpix=lcrpix;

  return(dv);
}
/*--------------------------------------------------------*/
void saveBF(int vopt, char *fname, int m, double dv, double crpix,
  gsl_vector *BF)
{
        char  *oname,
              **header,
              line[CARD_SIZE+1];  // +1 because of INAME header card
        int   i;
        float *fbf;
        FILE  *outf;

  if ((header=(char **)calloc(RECORD_CARDS, sizeof(char *))) == NULL)
  { perror("\n\tERROR! saveBF: calloc(header)"); exit(EXIT_FAILURE); }

  for (i=0; i<RECORD_CARDS; i++)
  {
    if ((header[i]=(char *)calloc(CARD_SIZE, sizeof(char))) == NULL)
    { perror("\n\tERROR! saveBF: calloc(header[i])"); exit(EXIT_FAILURE); }
    memset(header[i], ' ', CARD_SIZE);
  }

  sprintf(line, "SIMPLE  =                    T");
  memcpy(header[0], line, strlen(line));

  sprintf(line, "BITPIX  =                  -32 /");
  memcpy(header[1], line, strlen(line));

  sprintf(line, "NAXIS   =                    1 /");
  memcpy(header[2], line, strlen(line));

  sprintf(line, "NAXIS1  =               %6d /", m);
  memcpy(header[3], line, strlen(line));

  sprintf(line, "OBJECT  = 'Broadening Function' /");
  memcpy(header[4], line, strlen(line));

  sprintf(line, "INAME   = '%-68s'", fname);
  memcpy(header[5], line, strlen(line));

  sprintf(line, "CRVAL1  = %16.14e", -dv*(m/2+1.0-crpix));
  memcpy(header[6], line, strlen(line));

  sprintf(line, "CD1_1   = %16.14e", dv);
  memcpy(header[7], line, strlen(line));

  sprintf(line, "CDELT1  = %16.14e", dv);
  memcpy(header[8], line, strlen(line));

  sprintf(line, "CRPIX1  = %16.14e", crpix);
  memcpy(header[9], line, strlen(line));

  sprintf(line, "END       ");
  memcpy(header[10], line, strlen(line));

  if ((fbf=(float *)calloc(m, sizeof(float))) == NULL)
  { perror("\n\tERROR! saveBF: calloc(fbf)"); exit(EXIT_FAILURE); }

  for (i=0; i<m; i++) fbf[i]=(float)gsl_vector_get(BF, i);

  if (vopt > 0)
  {
    for (i=0; i<m; i++)
      printf("%d  %g fbf[%d]= %g\n", i-m/2, dv*(i-(m/2)), i, fbf[i]);
    printf("%g\n", -dv*(m/2));
  }

  if ((oname=(char *)calloc(strlen(fname)+4, sizeof(char))) == NULL)
  { perror("\n\tERROR! saveBF: calloc(oname)"); exit(EXIT_FAILURE); }

  strcpy(oname, "bf_"); strcat(oname, fname);
  if (vopt > 0) printf("\nOutput file:\n%s\n\n", oname);

  if ((outf=fopen(oname, "w")) == NULL) { perror(oname); exit(EXIT_FAILURE); }

  write_FITS_header(outf, RECORD_CARDS, header);
  write_FITS_1Dimage(outf, m, sizeof(float), fbf);

  fclose(outf);

  free(fbf);
  for (i=0; i<RECORD_CARDS; i++) free(header[i]);
  free(header);

  return;
}
/*--------------------------------------------------------*/
int main(int argc, char *argv[])
{
        char        *svdname,   // input svd file name
                    *sname,     // input spectrum file name
                    **header,   // FITS header
                    vopt;
        int         j,          // loop numerators
                    hsize,      // number of FITS header lines
                    m,          // length of the BF
                    n;          // size of U
        long        ns;         // length of the spectrum
        float       *data;      // spectrum data buffer
        double      dv,         // step in radial velocity (taken from W header)
                    crpix;
        gsl_vector  *S,     // S vector in SVD
                    *SSR,   // 1 - (truncated spectrum)
                    *BF;    // Broadening Function
        gsl_matrix  *U,     // U matrix in SVD
                    *V;     // VT matrix in SVD

  vopt=0;
  svdname = sname = NULL;
  switch (argc)
  {
    case  4:  if (strcasecmp(argv[3], "-v") == 0) vopt=1;
    case  3:  svdname=argv[1];
              sname=argv[2];
              break;
    default:  usage();
  }

  dv=read_svd(vopt, svdname, &m, &n, &S, &U, &V, &crpix);
  if (vopt > 0)
    printf("\n%s:  n= %d  m= %d  dv= %f [km/s]\n", svdname, n, m, dv);

  data=read_FITS_1Dfile(sname, &hsize, &header, &ns);
  for (j=0; j<hsize; j++) free(header[j]);
  free(header);
  if (vopt > 0) printf("%s: %ld points\n", sname, ns);

  if (ns != n+m-1)
  {
    printf("\n\tERROR! bf: Size of the spectrum doesn't agree with reference.\n");
    exit(EXIT_FAILURE);
  }

  SSR=gsl_vector_calloc(n);
  BF=gsl_vector_calloc(m);

  for (j=0; j<n; j++) gsl_vector_set(SSR, j, 1.0-data[j+m/2]);
  free(data);

  if (vopt > 0) printf("gsl_linalg_SV_solve...\n");
  gsl_linalg_SV_solve(U, V, S, SSR, BF);

  gsl_vector_free(S);
  gsl_vector_free(SSR);
  gsl_matrix_free(U);
  gsl_matrix_free(V);

  saveBF(vopt, sname, m, dv, crpix, BF);

  gsl_vector_free(BF);

  return(0);
}
/*** END ***/
