/*========================================================*/
/*                                                        */
/*  gsmooth.c             version 0.7.1     2010.01.15    */
/*                                                        */
/*  Copyright (C) 2010 by Wojtek Pych, CAMK PAN           */
/*  pych@camk.edu.pl                                      */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*                                                        */
/*  Gaussian Smoothing of Broadening Functions.           */
/*  Analog of BFpro3 by S. Rucinski in IDL.               */
/*                                                        */
/*========================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "pfitsio.h"

void usage(void);

/*--------------------------------------------------------*/
void usage()
{
  printf("\n\t USAGE:  gsmooth  input_file  S  output_file\n\n");
  printf("\tg = exp(-x/S*x/S)\n\n");
  printf("S in km/s\n");
  exit(EXIT_FAILURE);
}
/*--------------------------------------------------------*/
float *gaussprof(int n, float s)
{
        int     i,
                k;
        float  *y,
                a,
                b;

  k=n/2;

  if ((y=(float *)calloc(n, sizeof(float))) == NULL)
  { perror("\n\tERROR! gaussprof: calloc(y)"); exit(EXIT_FAILURE); }
  y+=k;

  a=sqrt(M_PI)*s;
  for (i=-k; i<=k; i++)
  {
    b=i/s;
    y[i]=exp(-b*b)/a;
  }

  return(y);
}
/*--------------------------------------------------------*/
float *convolve(int n1, float *y, int n2, float *g)
{
        int     i, j,
                k;
        float  *conv;

  k=n2/2;

  if ((conv=(float *)calloc(n1, sizeof(float))) == NULL)
  { perror("\n\tERROR! convolve: calloc(conv)"); exit(EXIT_FAILURE); }

  for (i=0; i<n1; i++)
  {
    conv[i]=0.0;
    for (j=-k; j<=k; j++)
    {
      if ((i+j >= 0) && (i+j < n1)) conv[i]+=y[i+j]*g[j];
    }
  }

  return(conv);
}
/*--------------------------------------------------------*/
void saveBF(char *oname, int *ncards, char ***header, float sigma, int m,
  float *bf)
{
        char  line[CARD_SIZE+1];
        int   p;
        FILE  *outf;

  memset(line, ' ', CARD_SIZE);
  sprintf(line, "HISTORY = 'Gaussian smoothing: sigma= %f km/s'", sigma);
  if (strlen(line) > CARD_SIZE) line[CARD_SIZE]='\0';

  if ((p=add_header_card(line, ncards, header)) == 0)
    printf("\n\tWARNING! saveBF(): add_header_card - failed\n\n");

  if ((outf=fopen(oname, "w")) == NULL) { perror(oname); exit(EXIT_FAILURE); }
  write_FITS_header(outf, *ncards, *header);
  write_FITS_1Dimage(outf, (long)m, sizeof(float), bf);
  fclose(outf);

  return;
}
/*--------------------------------------------------------*/
int main(int argc, char *argv[])
{
        char    *iname,
                *oname,
                **header,
                hline[CARD_SIZE];
        int     i,
                n,
                ng,
                ncards;
        long    ln;
        float   *data,
                *conv,
                *g,
                s, sm,
                dt;

  if (argc != 4) usage();

  iname=argv[1];
  s=atof(argv[2]);  // width of the Gaussian in km/s
  oname=argv[3];

  data=read_FITS_1Dfile(iname, &ncards, &header, &ln);
  n=(int)ln;
  printf("%s: %d data points\n", iname, n);

  if (get_FITS_key(ncards, header, "CD1_1", hline) < 1)
  {
    if (get_FITS_key(ncards, header, "CDELT1", hline) < 1)
    {
      printf("\n\tWARNING! no CD1_1 nor CDELT1 found in the FITS header: ");
      printf("using 1.0\n");
      dt=1.0;
    }
    else  sscanf(hline, "%g", &dt);
  }
  else  sscanf(hline, "%g", &dt);
   
  if (dt <= 0.0)
  {
    printf("\n\tERROR! gsmooth():  dt <= 0\n");
    printf("\t Check CD1_1 or CDELT1 in the FITS header.\n");
    exit(EXIT_FAILURE);
  }

  sm=s/dt;
  ng=2*(int)(sm*sqrt(-log(1.0e-8)))+1;
  printf("dV= %g  ng= %d\n", dt, ng);

  g=gaussprof(ng, sm);

  conv=convolve(n, data, ng, g);

  free(data);
  free(g-ng/2);

  saveBF(oname, &ncards, &header, s, n, conv);
  printf("Output file:\n %s\n\n", oname);

  free(conv);
  for (i=0; i<ncards; i++) free(header[i]);
  free(header);

  return(0);
}
/*** END ***/
