/*========================================================*/
/*                                                        */
/*  pfitsin.c       version 3.5.0         2010.01.13      */
/*                                                        */
/*  Copyright (C) 2010 by Wojtek Pych, CAMK PAN           */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*  Private FITS I/O library.                             */
/*                                                        */
/*  Definition of the FITS:                               */
/*  Hanisch, R. J.; et al. 2001, A&A, 376, 359-380        */
/*                                                        */
/*========================================================*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pfitsin.h"

/*--------------------------------------------------------*/
long read_FITS_image(FILE *inf, char **header, const int ncards,
  void **buf)
{
        void  *ptr;                     /* data table pointer   */
        char  keyword[KEYWORD_SIZE+1],  /* header entry keyword */
              value[VALUE_SIZE];
        int   i,              /* loop numerator               */
              naxes,          /* image dimension              */
              *naxis,         /* table of sizes of axes       */
              bitpix,         /* bits per pixel               */
              bytepix;        /* bytes per pixel              */
        long  pixnum;         /* number of image pixels       */

  memset(value, (int)' ', VALUE_SIZE);

  if (get_FITS_key(ncards, header, "NAXIS", value) == -1) return(0);
  if (sscanf(value, "%d", &naxes) < 1)
  {
    printf("\n\tERROR! read_FITS_image(): reading NAXIS\n");
    return(0);
  }

  if ((naxis=(int *)calloc((size_t)naxes, sizeof(int))) == NULL)
  { perror("\n\tERROR! read_FITS_image(): calloc(naxis)"); exit(EXIT_FAILURE); }

  for (i=0; i<naxes; i++)
  {
    if (snprintf(keyword, KEYWORD_SIZE, "NAXIS%d", i+1) < 6)
    {
      printf("\n\tERROR! read_FITS_image(): setting NAXIS_ keyword\n");
      free(naxis);
      return(0);
    }
    if (get_FITS_key(ncards, header, keyword, value) == -1)
    {
      free(naxis);
      return(0);
    }
    if (sscanf(value, "%d", &naxis[i]) < 1)
    {
      printf("\n\tERROR! read_FITS_image(): reading %s\n", keyword);
      free(naxis);
      return(0);
    }
  }

  pixnum=1;
  for (i=0; i<naxes; i++) pixnum*=naxis[i];

  free(naxis);

  if (get_FITS_key(ncards, header, "BITPIX", value) == -1) return(0);
  if (sscanf(value, "%d", &bitpix) < 1)
  {
    printf("\n\tERROR! read_FITS_image(): reading BITPIX from header\n");
    return(0);
  }
  bytepix=abs(bitpix)/8;

  if ((ptr=(void *)malloc(pixnum*bytepix)) == NULL)
  { perror("\n\tERROR! read_FITS_image(): malloc(ptr)"); exit(EXIT_FAILURE); }

  if (fread(ptr, (size_t)bytepix, pixnum, inf) != pixnum)
  {
    printf("\n\tERROR! read_FITS_image(): reading the image\n");
    return(0);
  }

  if (needswap() != 0) swap_bytes(ptr, pixnum, bytepix);

  *buf=ptr;

  return(pixnum);
}
/*--------------------------------------------------------*/
float *read_FITS_1Dfile(char *iname, int *ncards, char ***header, long *n1)
{
        void    *buf;         /* FITS image buffer        */
        char    value[VALUE_SIZE],  /* card value buffer  */
		**lhead;            /* pointer to header  */
        int     i,            /* loop numerator           */
		hs,           /* number of header cards   */
                bitpix,       /* FITS BITPIX value        */
                naxes;        /* number of image axes     */
        long    nx,
                nbytes;       /* length of the image buffer */
        float   *ldata;       /* output data pointer      */
        double  FITS_bzero,
                FITS_bscale;
        FILE    *inf;

  memset(value, (int)' ', VALUE_SIZE);

  if ((inf=fopen(iname, "r")) == NULL)
  {
    printf("\n\tERROR! read_FITS_1Dfile():"); perror(iname);
    exit(EXIT_FAILURE);
  }

  if ((hs=read_FITS_header(inf, &lhead)) < RECORD_CARDS)
  {
    printf("\n\tERROR! read_FITS_1Dfile(): incorrect header length\n");
    return(NULL);
  }

  if ((nbytes=read_FITS_image(inf, lhead, hs, &buf)) < 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(): incorrect image length\n");
    return(NULL);
  }

  if (fclose(inf) != 0)
    printf("\n\tWARNING! read_FITS_1Dfile: cannot close %s\n", iname);

  if (get_FITS_key(hs, lhead, "NAXIS", value) == -1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(%s): NAXIS not found in the header\n",
            iname);
    return(NULL);
  }

  if (sscanf(value, "%d", &naxes) < 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(): reading NAXIS\n");
    return(NULL);
  }

  if (naxes != 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(%s): NAXIS= %d (expected 1)\n",
            iname, naxes);
    return(NULL);
  }

  if (get_FITS_key(hs, lhead, "NAXIS1", value) == -1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(%s): NAXIS1 not found in the header\n",
            iname);
    return(NULL);
  }

  if (sscanf(value, "%ld", &nx) < 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(): reading NAXIS1\n");
    return(NULL);
  }

  if (get_FITS_key(hs, lhead, "BITPIX", value) == -1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(%s): BITPIX not found in the header\n",
            iname);
    return(NULL);
  }

  if (sscanf(value, "%d", &bitpix) < 1)
  {
    printf("\n\tERROR! read_FITS_1Dfile(): reading BITPIX\n"); return(NULL);
  }

  if ((ldata=(float *)calloc((size_t)nx, sizeof(long))) == NULL)
  {
    perror("\n\tERROR! read_FITS_1Dfile(): calloc(ldata)"); exit(EXIT_FAILURE);
  }

  if (get_FITS_key(hs, lhead, "BZERO", value) == -1)  FITS_bzero=0.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bzero) < 1)
    {
      printf("\n\tERROR! read_FITS_1Dfile(): reading BZERO\n");
      return(NULL);
    }
  }

  if (get_FITS_key(hs, lhead, "BSCALE", value) == -1) FITS_bscale=1.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bscale) < 1)
    {
      printf("\n\tERROR! read_FITS_1Dfile(): reading BSCALE\n");
      return(NULL);
    }
  }

  switch(bitpix)
  {
    case   8: for (i=0; i<nx; i++)  ldata[i]=(float)((char *)buf)[i];
              break;
    case  16: for (i=0; i<nx; i++)  ldata[i]=(float)((short *)buf)[i];
              break;
    case  32: for (i=0; i<nx; i++)  ldata[i]=(float)((long *)buf)[i];
              break;
    case -32: for (i=0; i<nx; i++)  ldata[i]=((float *)buf)[i];
              break;
    case -64: for (i=0; i<nx; i++)  ldata[i]=(float)((double *)buf)[i];
              break;
    default:  printf("\n\tERROR! read_FITS_1Dfile(%s): bitpix= %d",
                      iname, bitpix);
              printf(" not supported\n");
              return(NULL);
  }

  free(buf);

  if (fabs(FITS_bscale-1.0) > EPSILON)
    for (i=0; i<nx; i++) ldata[i]*=FITS_bscale;
  if (fabs(FITS_bzero) > EPSILON)
    for (i=0; i<nx; i++) ldata[i]+=FITS_bzero;

  *n1=nx;
  *ncards=hs;
  *header=lhead;

  return(ldata);
}
/*--------------------------------------------------------*/
float **read_FITS_2Dfile(char *iname, int *ncards, char ***header,
  long *n1, long *n2)
{
        void    *buf;
        char    value[VALUE_SIZE],
		**lhead;
        int     i,
                j,
		hs,
                bitpix,
                naxes;
        long    nx,
                ny,
                nbytes;
        float   **ldata;
        double  FITS_bzero,
                FITS_bscale;
        FILE    *inf;

  memset(value, (int)' ', VALUE_SIZE);

  if ((inf=fopen(iname, "r")) == NULL)
  { printf("\n\tERROR! "); perror(iname); exit(EXIT_FAILURE); }

  if ((hs=read_FITS_header(inf, &lhead)) < RECORD_CARDS)
  {
    printf("\n\tERROR! read_FITS_2Dfile(): incorrect header length\n");
    return(NULL);
  }

  if ((nbytes=read_FITS_image(inf, lhead, hs, &buf)) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(): incorrect image length\n");
    return(NULL);
  }

  if (fclose(inf) != 0)
    printf("\n\tWARNING! read_FITS_2Dfile(): cannot close %s\n", iname);

  if (get_FITS_key(hs, lhead, "NAXIS", value) == -1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(%s): NAXIS not found in the header\n",
            iname);
    return(NULL);
  }

  if (sscanf(value, "%d", &naxes) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(): reading NAXIS\n");
    return(NULL);
  }

  if (naxes != 2)
  {
    printf("\n\tERROR! read_FITS_2Dfile(%s): NAXIS= %d (expected 2)\n",
            iname, naxes);
    return(NULL);
  }

  if (get_FITS_key(hs, lhead, "NAXIS1", value) == -1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(%s): NAXIS1 not found in header\n",
            iname);
    return(NULL);
  }

  if (sscanf(value, "%ld", &nx) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(): reading NAXIS1\n");
    return(NULL);
  }

  if (get_FITS_key(hs, lhead, "NAXIS2", value) == -1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(%s): NAXIS2 not found in header\n",
            iname);
    return(NULL);
  }

  if (sscanf(value, "%ld", &ny) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(): reading NAXIS2\n");
    return(NULL);
  }

  if (get_FITS_key(hs, lhead, "BITPIX", value) == -1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(%s): BITPIX not found in header\n",
            iname);
    return(NULL);
  }

  if (sscanf(value, "%d", &bitpix) < 1)
  {
    printf("\n\tERROR! read_FITS_2Dfile(): reading BITPIX\n");
    return(NULL);
  }

  if (get_FITS_key(hs, lhead, "BZERO", value) == -1)  FITS_bzero=0.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bzero) < 1)
    {
      printf("\n\tERROR! read_FITS_2Dfile(): reading BZERO\n");
      return(NULL);
    }
  }

  if (get_FITS_key(hs, lhead, "BSCALE", value) == -1) FITS_bscale=1.0;
  else
  {
    if (sscanf(value, "%lg", &FITS_bscale) < 1)
    {
      printf("\n\tERROR! read_FITS_2Dfile: reading BSCALE\n");
      return(NULL);
    }
  }

  if ((ldata=(float **)calloc((size_t)ny, sizeof(float *))) == NULL)
  {
    perror("\n\tERROR! read_FITS_2Dfile(): calloc(ldata)"); exit(EXIT_FAILURE);
  }

  for (i=0; i<ny; i++)
  {
    if ((ldata[i]=(float *)calloc((size_t)nx, sizeof(float))) == NULL)
    {
      perror("\n\tERROR! read_FITS_2Dfile(): calloc(ldata[i])");
      exit(EXIT_FAILURE);
    }
  }

  switch(bitpix)
  {
    case   8: for (i=0; i<ny; i++)
                for (j=0; j<nx; j++)
                  ldata[i][j]=(float)((char *)buf)[i*nx+j];
              break;
    case  16: for (i=0; i<ny; i++)
                for (j=0; j<nx; j++)
                  ldata[i][j]=(float)((short *)buf)[i*nx+j];
              break;
    case  32: for (i=0; i<ny; i++)
                for (j=0; j<nx; j++)
                  ldata[i][j]=(float)((long *)buf)[i*nx+j];
              break;
    case -32: for (i=0; i<ny; i++)
                for (j=0; j<nx; j++)
                  ldata[i][j]=((float *)buf)[i*nx+j];
              break;
    case -64: for (i=0; i<ny; i++)
                for (j=0; j<nx; j++)
                  ldata[i][j]=(float)((double *)buf)[i*nx+j];
              break;
    default:  printf("\n\tERROR! %s: bitpix=%d not supported\n", iname, bitpix);
              exit(EXIT_FAILURE);
  }

  free(buf);

  if (fabs(FITS_bscale-1.0) > EPSILON)
    for (i=0; i<ny; i++)
      for (j=0; j<nx; j++)
        ldata[i][j]*=FITS_bscale;

  if (fabs(FITS_bzero) > EPSILON)
    for (i=0; i<ny; i++)
      for (j=0; j<nx; j++)
        ldata[i][j]+=FITS_bzero;

  *n1=nx;
  *n2=ny;
  *header=lhead;
  *ncards=hs;

  return(ldata);
}
/*--------------------------------------------------------*/
/*** END ***/
