/*========================================================*/
/*                                                        */
/*  template_svd.c        version 1.4.0    2010.01.15     */
/*                                                        */
/*  Copyright (C) 2010 by Wojtek Pych, CAMK PAN           */
/*                                                        */
/*  Written for GNU project C and C++ Compiler            */
/*                                                        */
/*  Singular Value Decomposition of the template spectrum */
/*  using GLS gsl_linalg_SV_decomp_mod() function.        */
/*                                                        */
/*========================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pfitsio.h"

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

#include <gsl/gsl_linalg.h>

#define STDSTRLEN 200

typedef struct PAR
{
  char  *iname,
        *oname;
  int   m,    // width of the BF
        vopt; // verbose level
} PAR;

void usage(void);
gsl_matrix *map4(int, float *, int);
char **primary_header(PAR, int, char **);
char **A_header(PAR, int, char **, int, int);
char **V_header(PAR, int, char **, int, int);
void write_FITS_2Dimage_d(FILE *, long, long, int, void **);
void saveSVD(PAR, int, char **, gsl_matrix *, gsl_matrix *, gsl_vector *);

/*--------------------------------------------------------*/
void usage()
{
  printf("\n\t USAGE: template_svd  template output_file N [-v]\n\n");
  printf("template_spectrum: name of the input FITS file\n");
  printf("output_file:       name of the output FITS file\n");
  printf("N:                 width of the Broadening Function (odd number)\n");
  printf("-v:                verbose\n\n");
  exit(1);
}
/*--------------------------------------------------------*/
gsl_matrix *map4(int n, float *data, int m)
{
        int         i, j;
        gsl_matrix  *des;

  for (i=0; i<n; i++) data[i]=1.0-data[i];

  des=gsl_matrix_calloc(n-m+1, m);

  for (i=m/2; i<n-m/2; i++)
    for (j=0; j<m; j++)
      gsl_matrix_set(des, i-m/2, j, data[i-j+m/2]);

  return(des);
}
/*--------------------------------------------------------*/
char **primary_header(PAR par, int ncards, char **header)
{
        char  **new_header,
              line[CARD_SIZE];
        short val, delt, crpix1;
        int   i, j;

  val = delt = crpix1 = 0;

  if ((new_header=(char **)calloc(RECORD_CARDS, sizeof(char *))) == NULL)
  {
    perror("\n\tERROR! primary_header(): calloc(new_header)");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<RECORD_CARDS; i++)
  {
    if ((new_header[i]=(char *)calloc(CARD_SIZE, sizeof(char))) == NULL)
    {
      perror("\n\tERROR! primary_header(): calloc(new_header[i])");
      exit(EXIT_FAILURE);
    }
    memset(new_header[i], ' ', CARD_SIZE);
  }

  sprintf(line, "SIMPLE  =                    T");
  memcpy(new_header[0], line, strlen(line));

  sprintf(line, "BITPIX  =                  -64 /");
  memcpy(new_header[1], line, strlen(line));

  sprintf(line, "NAXIS   =                    1 /");
  memcpy(new_header[2], line, strlen(line));

  sprintf(line, "NAXIS1  =               %6d /", par.m);
  memcpy(new_header[3], line, strlen(line));

  sprintf(line, "EXTEND  =                    T / Extensions will follow");
  memcpy(new_header[4], line, strlen(line));

  sprintf(line, "CONTENT =   'vector S (SVD)'   /");
  memcpy(new_header[5], line, strlen(line));

  sprintf(line, "INAME   = '%s'", par.iname);
  memcpy(new_header[6], line, strlen(line));

  j=7;
  line[KEYWORD_SIZE]='\0';
  for (i=0; i<ncards; i++)
  {
    memcpy(line, header[i], KEYWORD_SIZE);

    if (strcmp(line, "OBJECT  ") == 0)
    {
      memcpy(new_header[j], header[i], CARD_SIZE);
      j++;
    }
    else if (strcmp(line, "CRVAL1  ") == 0)
    {
      val++;
      memcpy(new_header[j], header[i], CARD_SIZE);
      memcpy(new_header[j], "O_CRVAL1", KEYWORD_SIZE);
      j++;
    }
    else if (strcmp(line, "CDELT1  ") == 0)
    {
      delt++;
      memcpy(new_header[j], header[i], CARD_SIZE);
      memcpy(new_header[j], "O_CDELT1", KEYWORD_SIZE);
      j++;
    }
    else if (strcmp(line, "CD1_1   ") == 0)
    {
      delt++;
      memcpy(new_header[j], header[i], CARD_SIZE);
      memcpy(new_header[j], "O_CD1_1 ", KEYWORD_SIZE);
      j++;
    }
    else if (strcmp(line, "CRPIX1  ") == 0)
    {
      crpix1++;
      memcpy(new_header[j], header[i], CARD_SIZE);
      memcpy(new_header[j], "O_CRPIX1", KEYWORD_SIZE);
      j++;
    }
  }
      
  sprintf(line, "END       ");
  memcpy(new_header[j], line, strlen(line));

  if (val == 0)
    printf("\n\tWARNING! CRVAL1 not found in the FITS header.\n");
  else if (val > 1)
    printf("\n\tWARNING! %d cards with CRVAL1 found in the FITS header.\n",
      val);

  if (delt == 0)
    printf("\n\tWARNING! no CDELT1 nor CD1_1 found in the FITS header.\n");
  else if (delt > 2)
    printf("\n\tWARNING! %d cards with CDELT1 or CD1_1 in the FITS header.\n",
      delt);

  if (crpix1 == 0)
    printf("\n\tWARNING! no CRPIX1 found in the FITS header.\n");
  else if (crpix1 > 2)
    printf("\n\tWARNING! %d cards with CRPIX1 found in the FITS header.\n",
      crpix1);

  return(new_header);
}
/*--------------------------------------------------------*/
char **A_header(PAR par, int ncards, char **header, int n1, int n2)
{
        char  **new_header,
              line[CARD_SIZE];
        int   i;

  if ((new_header=(char **)calloc(RECORD_CARDS, sizeof(char *))) == NULL)
  { perror("\n\tERROR! A_header(): calloc(new_header)"); exit(EXIT_FAILURE); }

  for (i=0; i<RECORD_CARDS; i++)
  {
    if ((new_header[i]=(char *)calloc(CARD_SIZE, sizeof(char))) == NULL)
    {
      perror("\n\tERROR! A_header(): calloc(new_header[i])");
      exit(EXIT_FAILURE);
    }
    memset(new_header[i], ' ', CARD_SIZE);
  }

  sprintf(line, "XTENSION= 'IMAGE   '");
  memcpy(new_header[0], line, strlen(line));

  sprintf(line, "BITPIX  =                  -64 /");
  memcpy(new_header[1], line, strlen(line));

  sprintf(line, "NAXIS   =                    2 /");
  memcpy(new_header[2], line, strlen(line));

  sprintf(line, "NAXIS1  =               %6d /", n1);
  memcpy(new_header[3], line, strlen(line));

  sprintf(line, "NAXIS2  =               %6d /", n2);
  memcpy(new_header[4], line, strlen(line));

  sprintf(line, "PCOUNT  = 0");
  memcpy(new_header[5], line, strlen(line));

  sprintf(line, "GCOUNT  = 1");
  memcpy(new_header[6], line, strlen(line));

  sprintf(line, "EXTNAME =   'matrix A (SVD)'   /");
  memcpy(new_header[7], line, strlen(line));

  sprintf(line, "END       ");
  memcpy(new_header[8], line, strlen(line));

  return(new_header);
}
/*--------------------------------------------------------*/
char **V_header(PAR par, int ncards, char **header, int n1, int n2)
{
        char  **new_header,
              line[CARD_SIZE];
        int   i;

  if ((new_header=(char **)calloc(RECORD_CARDS, sizeof(char *))) == NULL)
  { perror("\n\tERROR! V_header(): calloc(new_header)"); exit(EXIT_FAILURE); }

  for (i=0; i<RECORD_CARDS; i++)
  {
    if ((new_header[i]=(char *)calloc(CARD_SIZE, sizeof(char))) == NULL)
    {
      perror("\n\tERROR! V_header(): calloc(new_header[i])");
      exit(EXIT_FAILURE);
    }
    memset(new_header[i], ' ', CARD_SIZE);
  }

  sprintf(line, "XTENSION= 'IMAGE   '");
  memcpy(new_header[0], line, strlen(line));

  sprintf(line, "BITPIX  =                  -64 /");
  memcpy(new_header[1], line, strlen(line));

  sprintf(line, "NAXIS   =                    2 /");
  memcpy(new_header[2], line, strlen(line));

  sprintf(line, "NAXIS1  =               %6d /", n1);
  memcpy(new_header[3], line, strlen(line));

  sprintf(line, "NAXIS2  =               %6d /", n2);
  memcpy(new_header[4], line, strlen(line));

  sprintf(line, "PCOUNT  = 0");
  memcpy(new_header[5], line, strlen(line));

  sprintf(line, "GCOUNT  = 1");
  memcpy(new_header[6], line, strlen(line));

  sprintf(line, "EXTNAME =   'matrix V (SVD)'   /");
  memcpy(new_header[7], line, strlen(line));

  sprintf(line, "END       ");
  memcpy(new_header[8], line, strlen(line));

  return(new_header);
}
/*--------------------------------------------------------*/
void write_FITS_2Dimage_d(FILE *outf, long nx, long ny, int bytepix,
  void **data)
{
        char  *tmp;
        long  ls,
              size,
              nw;

  size=nx*ny;
  if (needswap()) swap_bytes(data, size, bytepix);

  if ((nw=fwrite(data, bytepix, size, outf)) != size)
  {
    printf("\n\tERROR! write_FITS_2Dimage_d(): fwrite(data) failed\n\n");
    exit(EXIT_FAILURE);
  }

  size*=bytepix;
  if ((ls=size%RECORD_SIZE) != 0) ls=RECORD_SIZE-ls;
  if ((tmp=(char *)malloc(ls)) == NULL)
  {
    perror("\n\tERROR! write_FITS_2Dimage_d(): malloc(tmp)"); exit(EXIT_FAILURE);
  }

  memset(tmp, ' ', ls);

  if ((nw=fwrite(tmp, 1, ls, outf)) != ls)
  {
    printf("\n\tERROR! write_FITS_2Dimage_d(): fwrite(tmp) failed\n\n");
    exit(EXIT_FAILURE);
  }

  free(tmp);

  return;
}
/*--------------------------------------------------------*/
void saveSVD(PAR par, int ncards, char **header,
             gsl_matrix *A, gsl_matrix *V, gsl_vector *S)
{
        char    **new_header;
        int     i;
        FILE    *outf;

  if (par.vopt > 0) printf("Writting: '%s'\n", par.oname);

  if ((outf=fopen(par.oname, "w")) == NULL)
  { perror(par.oname); exit(EXIT_FAILURE); }

  new_header=primary_header(par, ncards, header);
  write_FITS_header(outf, RECORD_CARDS, new_header);
  for (i=0; i<RECORD_CARDS; i++) free(new_header[i]);
  free(new_header);

  write_FITS_1Dimage(outf, par.m, sizeof(double), S->data);

  new_header=A_header(par, ncards, header, A->size1, A->size2);
  write_FITS_header(outf, RECORD_CARDS, new_header);
  for (i=0; i<RECORD_CARDS; i++) free(new_header[i]);
  free(new_header);
  write_FITS_2Dimage_d(outf, A->size1, A->size2, sizeof(double),
                      (void *)(A->data));

  new_header=V_header(par, ncards, header, V->size1, V->size2);
  write_FITS_header(outf, RECORD_CARDS, new_header);
  for (i=0; i<RECORD_CARDS; i++) free(new_header[i]);
  free(new_header);
  write_FITS_2Dimage_d(outf, V->size1, V->size2, sizeof(double),
                      (void *)V->data);

  fclose(outf);

  return;
}
/*--------------------------------------------------------*/
int main(int argc, char *argv[])
{
        char        **header;
        int         ncards,
                    i;
        long        n;
        float       *data;
        PAR         par;
        gsl_vector  *S,
                    *work;
        gsl_matrix  *A,
                    *V,
                    *X;

  if ((argc < 4) || (argc > 5)) usage();

  par.iname=argv[1];
  par.oname=argv[2];
  par.m=atoi(argv[3]);
  if (par.m < 3)
  { printf("N must be greater than 2\n"); exit(EXIT_FAILURE); }
  if ((par.m%2) == 0)
  { printf("N must be an odd number\n"); exit(EXIT_FAILURE); }

  par.vopt=0;
  if (argc == 5) if (strcasecmp(argv[4], "-v") == 0) par.vopt=1;

  if (par.vopt > 0)
  {
    printf("Input spectrum:                    %s\n", par.iname);
    printf("Output file with the SVD matrices: %s\n", par.oname);
    printf("Width of the broadening function:  %d\n", par.m);
    printf("-------------------------------------\n");
  }

  data=read_FITS_1Dfile(par.iname, &ncards, &header, &n);
  if (par.vopt > 0) printf("%s: %ld points\n", par.iname, n);

  if (par.vopt > 0) printf("map4...\n");
  A=map4(n, data, par.m);
  V=gsl_matrix_calloc(par.m, par.m);
  S=gsl_vector_calloc(par.m);

  X=gsl_matrix_calloc(par.m, par.m);
  work=gsl_vector_calloc(par.m);

  free(data);

  if (par.vopt > 0)
    printf("SV_decomp(%.2e elements)...\n", (double)(par.m*(n-par.m+1)));
  gsl_linalg_SV_decomp_mod(A, X, V, S, work);

  saveSVD(par, ncards, header, A, V, S);

  gsl_matrix_free(A);
  gsl_matrix_free(V);
  gsl_vector_free(S);

  gsl_matrix_free(X);
  gsl_vector_free(work);

  for (i=0; i<ncards; i++) free(header[i]);
  free(header);

  return(0);
}
/*** END ***/
